#!/usr/bin/python

# (c) 2016, NetApp, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1",
                    "status": ["preview"],
                    "supported_by": "community"}

DOCUMENTATION = """
---
module: netapp_e_storagepool
short_description: NetApp E-Series manage volume groups and disk pools
description: Create or remove volume groups and disk pools for NetApp E-series storage arrays.
version_added: '2.2'
author:
  - Kevin Hulquest (@hulquest)
  - Nathan Swartz (@ndswartz)
extends_documentation_fragment:
  - netapp.eseries
options:
  state:
    description:
      - Whether the specified storage pool should exist or not.
      - Note that removing a storage pool currently requires the removal of all defined volumes first.
    required: true
    choices: ["present", "absent"]
  name:
    description:
      - The name of the storage pool to manage
    required: true
  criteria_drive_count:
    description:
      - The number of disks to use for building the storage pool.
      - When I(state=="present") then I(criteria_drive_count) or I(criteria_min_usable_capacity) must be specified.
      - The pool will be expanded if this number exceeds the number of disks already in place (See expansion note below)
    required: false
    type: int
  criteria_min_usable_capacity:
    description:
      - The minimum size of the storage pool (in size_unit).
      - When I(state=="present") then I(criteria_drive_count) or I(criteria_min_usable_capacity) must be specified.
      - The pool will be expanded if this value exceeds its current size. (See expansion note below)
    required: false
    type: float
  criteria_drive_type:
    description:
      - The type of disk (hdd or ssd) to use when searching for candidates to use.
      - When not specified each drive type will be evaluated until successful drive candidates are found starting with
        the most prevalent drive type.
    required: false
    choices: ["hdd","ssd"]
  criteria_size_unit:
    description:
      - The unit used to interpret size parameters
    choices: ["bytes", "b", "kb", "mb", "gb", "tb", "pb", "eb", "zb", "yb"]
    default: "gb"
  criteria_drive_min_size:
    description:
      - The minimum individual drive size (in size_unit) to consider when choosing drives for the storage pool.
  criteria_drive_interface_type:
    description:
      - The interface type to use when selecting drives for the storage pool
      - If not provided then all interface types will be considered.
    choices: ["sas", "sas4k", "fibre", "fibre520b", "scsi", "sata", "pata"]
    required: false
  criteria_drive_require_da:
    description:
      - Ensures the storage pool will be created with only data assurance (DA) capable drives.
      - Only available for new storage pools; existing storage pools cannot be converted.
    default: false
    type: bool
    version_added: '2.9'
  criteria_drive_require_fde:
    description:
     - Whether full disk encryption ability is required for drives to be added to the storage pool
    default: false
    type: bool
  raid_level:
    description:
      - The RAID level of the storage pool to be created.
      - Required only when I(state=="present").
      - When I(raid_level=="raidDiskPool") then I(criteria_drive_count >= 10 or criteria_drive_count >= 11) is required
        depending on the storage array specifications.
      - When I(raid_level=="raid0") then I(1<=criteria_drive_count) is required.
      - When I(raid_level=="raid1") then I(2<=criteria_drive_count) is required.
      - When I(raid_level=="raid3") then I(3<=criteria_drive_count<=30) is required.
      - When I(raid_level=="raid5") then I(3<=criteria_drive_count<=30) is required.
      - When I(raid_level=="raid6") then I(5<=criteria_drive_count<=30) is required.
      - Note that raidAll will be treated as raidDiskPool and raid3 as raid5.
    required: false
    choices: ["raidAll", "raid0", "raid1", "raid3", "raid5", "raid6", "raidDiskPool"]
    default: "raidDiskPool"
  secure_pool:
    description:
      - Enables security at rest feature on the storage pool.
      - Will only work if all drives in the pool are security capable (FDE, FIPS, or mix)
      - Warning, once security is enabled it is impossible to disable without erasing the drives.
    required: false
    type: bool
  reserve_drive_count:
    description:
      - Set the number of drives reserved by the storage pool for reconstruction operations.
      - Only valid on raid disk pools.
    required: false
  remove_volumes:
    description:
    - Prior to removing a storage pool, delete all volumes in the pool.
    default: true
  erase_secured_drives:
    description:
      - If I(state=="absent") then all storage pool drives will be erase
      - If I(state=="present") then delete all available storage array drives that have security enabled.
    default: true
    type: bool
notes:
  - The expansion operations are non-blocking due to the time consuming nature of expanding volume groups
  - Traditional volume groups (raid0, raid1, raid5, raid6) are performed in steps dictated by the storage array. Each
    required step will be attempted until the request fails which is likely because of the required expansion time.
  - raidUnsupported will be treated as raid0, raidAll as raidDiskPool and raid3 as raid5.
  - Tray loss protection and drawer loss protection will be chosen if at all possible.
"""
EXAMPLES = """
- name: No disk groups
  netapp_e_storagepool:
    ssid: "{{ ssid }}"
    name: "{{ item }}"
    state: absent
    api_url: "{{ netapp_api_url }}"
    api_username: "{{ netapp_api_username }}"
    api_password: "{{ netapp_api_password }}"
    validate_certs: "{{ netapp_api_validate_certs }}"
"""
RETURN = """
msg:
    description: Success message
    returned: success
    type: str
    sample: Json facts for the pool that was created.
"""
import functools
from itertools import groupby
from time import sleep
from pprint import pformat
from ansible.module_utils.netapp import NetAppESeriesModule
from ansible.module_utils._text import to_native


def get_most_common_elements(iterator):
    """Returns a generator containing a descending list of most common elements."""
    if not isinstance(iterator, list):
        raise TypeError("iterator must be a list.")

    grouped = [(key, len(list(group))) for key, group in groupby(sorted(iterator))]
    return sorted(grouped, key=lambda x: x[1], reverse=True)


def memoize(func):
    """Generic memoizer for any function with any number of arguments including zero."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        class MemoizeFuncArgs(dict):
            def __missing__(self, _key):
                self[_key] = func(*args, **kwargs)
                return self[_key]

        key = str((args, kwargs)) if args and kwargs else "no_argument_response"
        return MemoizeFuncArgs().__getitem__(key)

    return wrapper


class NetAppESeriesStoragePool(NetAppESeriesModule):
    EXPANSION_TIMEOUT_SEC = 10
    DEFAULT_DISK_POOL_MINIMUM_DISK_COUNT = 11

    def __init__(self):
        version = "02.00.0000.0000"
        ansible_options = dict(
            state=dict(required=True, choices=["present", "absent"], type="str"),
            name=dict(required=True, type="str"),
            criteria_size_unit=dict(choices=["bytes", "b", "kb", "mb", "gb", "tb", "pb", "eb", "zb", "yb"],
                                    default="gb", type="str"),
            criteria_drive_count=dict(type="int"),
            criteria_drive_interface_type=dict(choices=["sas", "sas4k", "fibre", "fibre520b", "scsi", "sata", "pata"],
                                               type="str"),
            criteria_drive_type=dict(choices=["ssd", "hdd"], type="str", required=False),
            criteria_drive_min_size=dict(type="float"),
            criteria_drive_require_da=dict(type="bool", required=False),
            criteria_drive_require_fde=dict(type="bool", required=False),
            criteria_min_usable_capacity=dict(type="float"),
            raid_level=dict(choices=["raidAll", "raid0", "raid1", "raid3", "raid5", "raid6", "raidDiskPool"],
                            default="raidDiskPool"),
            erase_secured_drives=dict(type="bool", default=True),
            secure_pool=dict(type="bool", default=False),
            reserve_drive_count=dict(type="int"),
            remove_volumes=dict(type="bool", default=True))

        required_if = [["state", "present", ["raid_level"]]]
        super(NetAppESeriesStoragePool, self).__init__(ansible_options=ansible_options,
                                                       web_services_version=version,
                                                       supports_check_mode=True,
                                                       required_if=required_if)

        args = self.module.params
        self.state = args["state"]
        self.ssid = args["ssid"]
        self.name = args["name"]
        self.criteria_drive_count = args["criteria_drive_count"]
        self.criteria_min_usable_capacity = args["criteria_min_usable_capacity"]
        self.criteria_size_unit = args["criteria_size_unit"]
        self.criteria_drive_min_size = args["criteria_drive_min_size"]
        self.criteria_drive_type = args["criteria_drive_type"]
        self.criteria_drive_interface_type = args["criteria_drive_interface_type"]
        self.criteria_drive_require_fde = args["criteria_drive_require_fde"]
        self.criteria_drive_require_da = args["criteria_drive_require_da"]
        self.raid_level = args["raid_level"]
        self.erase_secured_drives = args["erase_secured_drives"]
        self.secure_pool = args["secure_pool"]
        self.reserve_drive_count = args["reserve_drive_count"]
        self.remove_volumes = args["remove_volumes"]
        self.pool_detail = None

        # Change all sizes to be measured in bytes
        if self.criteria_min_usable_capacity:
            self.criteria_min_usable_capacity = int(self.criteria_min_usable_capacity *
                                                    self.SIZE_UNIT_MAP[self.criteria_size_unit])
        if self.criteria_drive_min_size:
            self.criteria_drive_min_size = int(self.criteria_drive_min_size *
                                               self.SIZE_UNIT_MAP[self.criteria_size_unit])
        self.criteria_size_unit = "bytes"

        # Adjust unused raid level option to reflect documentation
        if self.raid_level == "raidAll":
            self.raid_level = "raidDiskPool"
        if self.raid_level == "raid3":
            self.raid_level = "raid5"

    @property
    @memoize
    def available_drives(self):
        """Determine the list of available drives"""
        return [drive["id"] for drive in self.drives if drive["available"] and drive["status"] == "optimal"]

    @property
    @memoize
    def available_drive_types(self):
        """Determine the types of available drives sorted by the most common first."""
        types = [drive["driveMediaType"] for drive in self.drives]
        return [entry[0] for entry in get_most_common_elements(types)]

    @property
    @memoize
    def available_drive_interface_types(self):
        """Determine the types of available drives."""
        interfaces = [drive["phyDriveType"] for drive in self.drives]
        return [entry[0] for entry in get_most_common_elements(interfaces)]

    @property
    def storage_pool_drives(self, exclude_hotspares=True):
        """Retrieve list of drives found in storage pool."""
        if exclude_hotspares:
            return [drive for drive in self.drives
                    if drive["currentVolumeGroupRef"] == self.pool_detail["id"] and not drive["hotSpare"]]

        return [drive for drive in self.drives if drive["currentVolumeGroupRef"] == self.pool_detail["id"]]

    @property
    def expandable_drive_count(self):
        """Maximum number of drives that a storage pool can be expanded at a given time."""
        capabilities = None
        if self.raid_level == "raidDiskPool":
            return len(self.available_drives)

        try:
            rc, capabilities = self.request("storage-systems/%s/capabilities" % self.ssid)
        except Exception as error:
            self.module.fail_json(msg="Failed to fetch maximum expandable drive count. Array id [%s].  Error[%s]."
                                      % (self.ssid, to_native(error)))

        return capabilities["featureParameters"]["maxDCEDrives"]

    @property
    def disk_pool_drive_minimum(self):
        """Provide the storage array's minimum disk pool drive count."""
        rc, attr = self.request("storage-systems/%s/symbol/getSystemAttributeDefaults" % self.ssid, ignore_errors=True)

        # Standard minimum is 11 drives but some allow 10 drives. 10 will be the default
        if (rc != 200 or "minimumDriveCount" not in attr["defaults"]["diskPoolDefaultAttributes"].keys() or
                attr["defaults"]["diskPoolDefaultAttributes"]["minimumDriveCount"] == 0):
            return self.DEFAULT_DISK_POOL_MINIMUM_DISK_COUNT

        return attr["defaults"]["diskPoolDefaultAttributes"]["minimumDriveCount"]

    def get_available_drive_capacities(self, drive_id_list=None):
        """Determine the list of available drive capacities."""
        if drive_id_list:
            available_drive_capacities = set([int(drive["usableCapacity"]) for drive in self.drives
                                              if drive["id"] in drive_id_list and drive["available"] and
                                              drive["status"] == "optimal"])
        else:
            available_drive_capacities = set([int(drive["usableCapacity"]) for drive in self.drives
                                              if drive["available"] and drive["status"] == "optimal"])

        self.module.log("available drive capacities: %s" % available_drive_capacities)
        return list(available_drive_capacities)

    @property
    def drives(self):
        """Retrieve list of drives found in storage pool."""
        drives = None
        try:
            rc, drives = self.request("storage-systems/%s/drives" % self.ssid)
        except Exception as error:
            self.module.fail_json(msg="Failed to fetch disk drives. Array id [%s].  Error[%s]."
                                      % (self.ssid, to_native(error)))

        return drives

    def is_drive_count_valid(self, drive_count):
        """Validate drive count criteria is met."""
        if self.criteria_drive_count and drive_count < self.criteria_drive_count:
            return False

        if self.raid_level == "raidDiskPool":
            return drive_count >= self.disk_pool_drive_minimum
        if self.raid_level == "raid0":
            return drive_count > 0
        if self.raid_level == "raid1":
            return drive_count >= 2 and (drive_count % 2) == 0
        if self.raid_level in ["raid3", "raid5"]:
            return 3 <= drive_count <= 30
        if self.raid_level == "raid6":
            return 5 <= drive_count <= 30
        return False

    @property
    def storage_pool(self):
        """Retrieve storage pool information."""
        storage_pools_resp = None
        try:
            rc, storage_pools_resp = self.request("storage-systems/%s/storage-pools" % self.ssid)
        except Exception as err:
            self.module.fail_json(msg="Failed to get storage pools. Array id [%s]. Error[%s]. State[%s]."
                                      % (self.ssid, to_native(err), self.state))

        pool_detail = [pool for pool in storage_pools_resp if pool["name"] == self.name]
        return pool_detail[0] if pool_detail else dict()

    @property
    def storage_pool_volumes(self):
        """Retrieve list of volumes associated with storage pool."""
        volumes_resp = None
        try:
            rc, volumes_resp = self.request("storage-systems/%s/volumes" % self.ssid)
        except Exception as err:
            self.module.fail_json(msg="Failed to get storage pools. Array id [%s]. Error[%s]. State[%s]."
                                      % (self.ssid, to_native(err), self.state))

        group_ref = self.storage_pool["volumeGroupRef"]
        storage_pool_volume_list = [volume["id"] for volume in volumes_resp if volume["volumeGroupRef"] == group_ref]
        return storage_pool_volume_list

    def get_ddp_capacity(self, expansion_drive_list):
        """Return the total usable capacity based on the additional drives."""

        def get_ddp_error_percent(_drive_count, _extent_count):
            """Determine the space reserved for reconstruction"""
            if _drive_count <= 36:
                if _extent_count <= 600:
                    return 0.40
                elif _extent_count <= 1400:
                    return 0.35
                elif _extent_count <= 6200:
                    return 0.20
                elif _extent_count <= 50000:
                    return 0.15
            elif _drive_count <= 64:
                if _extent_count <= 600:
                    return 0.20
                elif _extent_count <= 1400:
                    return 0.15
                elif _extent_count <= 6200:
                    return 0.10
                elif _extent_count <= 50000:
                    return 0.05
            elif _drive_count <= 480:
                if _extent_count <= 600:
                    return 0.20
                elif _extent_count <= 1400:
                    return 0.15
                elif _extent_count <= 6200:
                    return 0.10
                elif _extent_count <= 50000:
                    return 0.05

            self.module.fail_json(msg="Drive count exceeded the error percent table. Array[%s]" % self.ssid)

        def get_ddp_reserved_drive_count(_disk_count):
            """Determine the number of reserved drive."""
            reserve_count = 0

            if self.reserve_drive_count:
                reserve_count = self.reserve_drive_count
            elif _disk_count >= 256:
                reserve_count = 8
            elif _disk_count >= 192:
                reserve_count = 7
            elif _disk_count >= 128:
                reserve_count = 6
            elif _disk_count >= 64:
                reserve_count = 4
            elif _disk_count >= 32:
                reserve_count = 3
            elif _disk_count >= 12:
                reserve_count = 2
            elif _disk_count == 11:
                reserve_count = 1

            return reserve_count

        if self.pool_detail:
            drive_count = len(self.storage_pool_drives) + len(expansion_drive_list)
        else:
            drive_count = len(expansion_drive_list)

        drive_usable_capacity = min(min(self.get_available_drive_capacities()),
                                    min(self.get_available_drive_capacities(expansion_drive_list)))
        drive_data_extents = ((drive_usable_capacity - 8053063680) / 536870912)
        maximum_stripe_count = (drive_count * drive_data_extents) / 10

        error_percent = get_ddp_error_percent(drive_count, drive_data_extents)
        error_overhead = (drive_count * drive_data_extents / 10 * error_percent + 10) / 10

        total_stripe_count = maximum_stripe_count - error_overhead
        stripe_count_per_drive = total_stripe_count / drive_count
        reserved_stripe_count = get_ddp_reserved_drive_count(drive_count) * stripe_count_per_drive
        available_stripe_count = total_stripe_count - reserved_stripe_count

        return available_stripe_count * 4294967296

    @memoize
    def get_candidate_drives(self):
        """Retrieve set of drives candidates for creating a new storage pool."""

        def get_candidate_drive_request():
            """Perform request for new volume creation."""
            candidates_list = list()
            drive_types = [self.criteria_drive_type] if self.criteria_drive_type else self.available_drive_types
            interface_types = [self.criteria_drive_interface_type] \
                if self.criteria_drive_interface_type else self.available_drive_interface_types

            for interface_type in interface_types:
                for drive_type in drive_types:
                    candidates = None
                    volume_candidate_request_data = dict(
                        type="diskPool" if self.raid_level == "raidDiskPool" else "traditional",
                        diskPoolVolumeCandidateRequestData=dict(
                            reconstructionReservedDriveCount=65535))
                    candidate_selection_type = dict(
                        candidateSelectionType="count",
                        driveRefList=dict(driveRef=self.available_drives))
                    criteria = dict(raidLevel=self.raid_level,
                                    phyDriveType=interface_type,
                                    dssPreallocEnabled=False,
                                    securityType="capable" if self.criteria_drive_require_fde else "none",
                                    driveMediaType=drive_type,
                                    onlyProtectionInformationCapable=True if self.criteria_drive_require_da else False,
                                    volumeCandidateRequestData=volume_candidate_request_data,
                                    allocateReserveSpace=False,
                                    securityLevel="fde" if self.criteria_drive_require_fde else "none",
                                    candidateSelectionType=candidate_selection_type)

                    try:
                        rc, candidates = self.request("storage-systems/%s/symbol/getVolumeCandidates?verboseError"
                                                      "Response=true" % self.ssid, data=criteria, method="POST")
                    except Exception as error:
                        self.module.fail_json(msg="Failed to retrieve volume candidates. Array [%s]. Error [%s]."
                                                  % (self.ssid, to_native(error)))

                    if candidates:
                        candidates_list.extend(candidates["volumeCandidate"])

            # Sort output based on tray and then drawer protection first
            tray_drawer_protection = list()
            tray_protection = list()
            drawer_protection = list()
            no_protection = list()
            sorted_candidates = list()
            for item in candidates_list:
                if item["trayLossProtection"]:
                    if item["drawerLossProtection"]:
                        tray_drawer_protection.append(item)
                    else:
                        tray_protection.append(item)
                elif item["drawerLossProtection"]:
                    drawer_protection.append(item)
                else:
                    no_protection.append(item)

            if tray_drawer_protection:
                sorted_candidates.extend(tray_drawer_protection)
            if tray_protection:
                sorted_candidates.extend(tray_protection)
            if drawer_protection:
                sorted_candidates.extend(drawer_protection)
            if no_protection:
                sorted_candidates.extend(no_protection)

            return sorted_candidates

        # Determine the appropriate candidate list
        for candidate in get_candidate_drive_request():

            # Evaluate candidates for required drive count, collective drive usable capacity and minimum drive size
            if self.criteria_drive_count:
                if self.criteria_drive_count != int(candidate["driveCount"]):
                    continue
            if self.criteria_min_usable_capacity:
                if ((self.raid_level == "raidDiskPool" and self.criteria_min_usable_capacity >
                     self.get_ddp_capacity(candidate["driveRefList"]["driveRef"])) or
                        self.criteria_min_usable_capacity > int(candidate["usableSize"])):
                    continue
            if self.criteria_drive_min_size:
                if self.criteria_drive_min_size > min(self.get_available_drive_capacities(candidate["driveRefList"]["driveRef"])):
                    continue

            return candidate

        self.module.fail_json(msg="Not enough drives to meet the specified criteria. Array [%s]." % self.ssid)

    @memoize
    def get_expansion_candidate_drives(self):
        """Retrieve required expansion drive list.

        Note: To satisfy the expansion criteria each item in the candidate list must added specified group since there
        is a potential limitation on how many drives can be incorporated at a time.
            * Traditional raid volume groups must be added two drives maximum at a time. No limits on raid disk pools.

        :return list(candidate): list of candidate structures from the getVolumeGroupExpansionCandidates symbol endpoint
        """

        def get_expansion_candidate_drive_request():
            """Perform the request for expanding existing volume groups or disk pools.

            Note: the list of candidate structures do not necessarily produce candidates that meet all criteria.
            """
            candidates_list = None
            url = "storage-systems/%s/symbol/getVolumeGroupExpansionCandidates?verboseErrorResponse=true" % self.ssid
            if self.raid_level == "raidDiskPool":
                url = "storage-systems/%s/symbol/getDiskPoolExpansionCandidates?verboseErrorResponse=true" % self.ssid

            try:
                rc, candidates_list = self.request(url, method="POST", data=self.pool_detail["id"])
            except Exception as error:
                self.module.fail_json(msg="Failed to retrieve volume candidates. Array [%s]. Error [%s]."
                                          % (self.ssid, to_native(error)))

            return candidates_list["candidates"]

        required_candidate_list = list()
        required_additional_drives = 0
        required_additional_capacity = 0
        total_required_capacity = 0

        # determine whether and how much expansion is need to satisfy the specified criteria
        if self.criteria_min_usable_capacity:
            total_required_capacity = self.criteria_min_usable_capacity
            required_additional_capacity = self.criteria_min_usable_capacity - int(self.pool_detail["totalRaidedSpace"])

        if self.criteria_drive_count:
            required_additional_drives = self.criteria_drive_count - len(self.storage_pool_drives)

        # Determine the appropriate expansion candidate list
        if required_additional_drives > 0 or required_additional_capacity > 0:
            for candidate in get_expansion_candidate_drive_request():

                if self.criteria_drive_min_size:
                    if self.criteria_drive_min_size > min(self.get_available_drive_capacities(candidate["drives"])):
                        continue

                if self.raid_level == "raidDiskPool":
                    if (len(candidate["drives"]) >= required_additional_drives and
                            self.get_ddp_capacity(candidate["drives"]) >= total_required_capacity):
                        required_candidate_list.append(candidate)
                        break
                else:
                    required_additional_drives -= len(candidate["drives"])
                    required_additional_capacity -= int(candidate["usableCapacity"])
                    required_candidate_list.append(candidate)

                # Determine if required drives and capacities are satisfied
                if required_additional_drives <= 0 and required_additional_capacity <= 0:
                    break
            else:
                self.module.fail_json(msg="Not enough drives to meet the specified criteria. Array [%s]." % self.ssid)

        return required_candidate_list

    def get_reserve_drive_count(self):
        """Retrieve the current number of reserve drives for raidDiskPool (Only for raidDiskPool)."""

        if not self.pool_detail:
            self.module.fail_json(msg="The storage pool must exist. Array [%s]." % self.ssid)

        if self.raid_level != "raidDiskPool":
            self.module.fail_json(msg="The storage pool must be a raidDiskPool. Pool [%s]. Array [%s]."
                                      % (self.pool_detail["id"], self.ssid))

        return self.pool_detail["volumeGroupData"]["diskPoolData"]["reconstructionReservedDriveCount"]

    def get_maximum_reserve_drive_count(self):
        """Retrieve the maximum number of reserve drives for storage pool (Only for raidDiskPool)."""
        if self.raid_level != "raidDiskPool":
            self.module.fail_json(msg="The storage pool must be a raidDiskPool. Pool [%s]. Array [%s]."
                                      % (self.pool_detail["id"], self.ssid))

        drives_ids = list()

        if self.pool_detail:
            drives_ids.extend(self.storage_pool_drives)
            for candidate in self.get_expansion_candidate_drives():
                drives_ids.extend((candidate["drives"]))
        else:
            candidate = self.get_candidate_drives()
            drives_ids.extend(candidate["driveRefList"]["driveRef"])

        drive_count = len(drives_ids)
        maximum_reserve_drive_count = min(int(drive_count * 0.2 + 1), drive_count - 10)
        if maximum_reserve_drive_count > 10:
            maximum_reserve_drive_count = 10

        return maximum_reserve_drive_count

    def set_reserve_drive_count(self, check_mode=False):
        """Set the reserve drive count for raidDiskPool."""
        changed = False

        if self.raid_level == "raidDiskPool" and self.reserve_drive_count:
            maximum_count = self.get_maximum_reserve_drive_count()

            if self.reserve_drive_count < 0 or self.reserve_drive_count > maximum_count:
                self.module.fail_json(msg="Supplied reserve drive count is invalid or exceeds the maximum allowed. "
                                          "Note that it may be necessary to wait for expansion operations to complete "
                                          "before the adjusting the reserve drive count. Maximum [%s]. Array [%s]."
                                          % (maximum_count, self.ssid))

            if self.reserve_drive_count != self.get_reserve_drive_count():
                changed = True

            if not check_mode:
                try:
                    rc, resp = self.request("storage-systems/%s/symbol/setDiskPoolReservedDriveCount" % self.ssid,
                                            method="POST", data=dict(volumeGroupRef=self.pool_detail["id"],
                                                                     newDriveCount=self.reserve_drive_count))
                except Exception as error:
                    self.module.fail_json(msg="Failed to set reserve drive count for disk pool. Disk Pool [%s]."
                                              " Array [%s]." % (self.pool_detail["id"], self.ssid))

        return changed

    def erase_all_available_secured_drives(self, check_mode=False):
        """Erase all available drives that have encryption at rest feature enabled."""
        changed = False
        drives_list = list()
        for drive in self.drives:
            if drive["available"] and drive["fdeEnabled"]:
                changed = True
                drives_list.append(drive["id"])

        if drives_list and not check_mode:
            try:
                rc, resp = self.request("storage-systems/%s/symbol/reprovisionDrive?verboseErrorResponse=true"
                                        % self.ssid, method="POST", data=dict(driveRef=drives_list))
            except Exception as error:
                self.module.fail_json(msg="Failed to erase all secured drives. Array [%s]" % self.ssid)

        return changed

    def create_storage_pool(self):
        """Create new storage pool."""
        url = "storage-systems/%s/symbol/createVolumeGroup?verboseErrorResponse=true" % self.ssid
        request_body = dict(label=self.name,
                            candidate=self.get_candidate_drives())

        if self.raid_level == "raidDiskPool":
            url = "storage-systems/%s/symbol/createDiskPool?verboseErrorResponse=true" % self.ssid

            request_body.update(
                dict(backgroundOperationPriority="useDefault",
                     criticalReconstructPriority="useDefault",
                     degradedReconstructPriority="useDefault",
                     poolUtilizationCriticalThreshold=65535,
                     poolUtilizationWarningThreshold=0))

            if self.reserve_drive_count:
                request_body.update(dict(volumeCandidateData=dict(
                    diskPoolVolumeCandidateData=dict(reconstructionReservedDriveCount=self.reserve_drive_count))))

        try:
            rc, resp = self.request(url, method="POST", data=request_body)
        except Exception as error:
            self.module.fail_json(msg="Failed to create storage pool. Array id [%s].  Error[%s]."
                                      % (self.ssid, to_native(error)))

        # Update drive and storage pool information
        self.pool_detail = self.storage_pool

    def delete_storage_pool(self):
        """Delete storage pool."""
        storage_pool_drives = [drive["id"] for drive in self.storage_pool_drives if drive["fdeEnabled"]]
        try:
            delete_volumes_parameter = "?delete-volumes=true" if self.remove_volumes else ""
            rc, resp = self.request("storage-systems/%s/storage-pools/%s%s"
                                    % (self.ssid, self.pool_detail["id"], delete_volumes_parameter), method="DELETE")
        except Exception as error:
            self.module.fail_json(msg="Failed to delete storage pool. Pool id [%s]. Array id [%s].  Error[%s]."
                                      % (self.pool_detail["id"], self.ssid, to_native(error)))

        if storage_pool_drives and self.erase_secured_drives:
            try:
                rc, resp = self.request("storage-systems/%s/symbol/reprovisionDrive?verboseErrorResponse=true"
                                        % self.ssid, method="POST", data=dict(driveRef=storage_pool_drives))
            except Exception as error:
                self.module.fail_json(msg="Failed to erase drives prior to creating new storage pool. Array [%s]."
                                          " Error [%s]." % (self.ssid, to_native(error)))

    def secure_storage_pool(self, check_mode=False):
        """Enable security on an existing storage pool"""
        self.pool_detail = self.storage_pool
        needs_secure_pool = False

        if not self.secure_pool and self.pool_detail["securityType"] == "enabled":
            self.module.fail_json(msg="It is not possible to disable storage pool security! See array documentation.")
        if self.secure_pool and self.pool_detail["securityType"] != "enabled":
            needs_secure_pool = True

        if needs_secure_pool and not check_mode:
            try:
                rc, resp = self.request("storage-systems/%s/storage-pools/%s" % (self.ssid, self.pool_detail["id"]),
                                        data=dict(securePool=True), method="POST")
            except Exception as error:
                self.module.fail_json(msg="Failed to secure storage pool. Pool id [%s]. Array [%s]. Error"
                                          " [%s]." % (self.pool_detail["id"], self.ssid, to_native(error)))

        self.pool_detail = self.storage_pool
        return needs_secure_pool

    def migrate_raid_level(self, check_mode=False):
        """Request storage pool raid level migration."""
        needs_migration = self.raid_level != self.pool_detail["raidLevel"]
        if needs_migration and self.pool_detail["raidLevel"] == "raidDiskPool":
            self.module.fail_json(msg="Raid level cannot be changed for disk pools")

        if needs_migration and not check_mode:
            sp_raid_migrate_req = dict(raidLevel=self.raid_level)

            try:
                rc, resp = self.request("storage-systems/%s/storage-pools/%s/raid-type-migration"
                                        % (self.ssid, self.name), data=sp_raid_migrate_req, method="POST")
            except Exception as error:
                self.module.fail_json(msg="Failed to change the raid level of storage pool. Array id [%s]."
                                          "  Error[%s]." % (self.ssid, to_native(error)))

        self.pool_detail = self.storage_pool
        return needs_migration

    def expand_storage_pool(self, check_mode=False):
        """Add drives to existing storage pool.

        :return bool: whether drives were required to be added to satisfy the specified criteria."""
        expansion_candidate_list = self.get_expansion_candidate_drives()
        changed_required = bool(expansion_candidate_list)
        estimated_completion_time = 0.0

        # build expandable groupings of traditional raid candidate
        required_expansion_candidate_list = list()
        while expansion_candidate_list:
            subset = list()
            while expansion_candidate_list and len(subset) < self.expandable_drive_count:
                subset.extend(expansion_candidate_list.pop()["drives"])
            required_expansion_candidate_list.append(subset)

        if required_expansion_candidate_list and not check_mode:
            url = "storage-systems/%s/symbol/startVolumeGroupExpansion?verboseErrorResponse=true" % self.ssid
            if self.raid_level == "raidDiskPool":
                url = "storage-systems/%s/symbol/startDiskPoolExpansion?verboseErrorResponse=true" % self.ssid

            while required_expansion_candidate_list:
                candidate_drives_list = required_expansion_candidate_list.pop()
                request_body = dict(volumeGroupRef=self.pool_detail["volumeGroupRef"],
                                    driveRef=candidate_drives_list)
                try:
                    rc, resp = self.request(url, method="POST", data=request_body)
                except Exception as error:
                    rc, actions_resp = self.request("storage-systems/%s/storage-pools/%s/action-progress"
                                                    % (self.ssid, self.pool_detail["id"]), ignore_errors=True)
                    if rc == 200 and actions_resp:
                        actions = [action["currentAction"] for action in actions_resp
                                   if action["volumeRef"] in self.storage_pool_volumes]
                        self.module.fail_json(msg="Failed to add drives to the storage pool possibly because of actions"
                                                  " in progress. Actions [%s]. Pool id [%s]. Array id [%s]. Error[%s]."
                                                  % (", ".join(actions), self.pool_detail["id"], self.ssid,
                                                     to_native(error)))

                    self.module.fail_json(msg="Failed to add drives to storage pool. Pool id [%s]. Array id [%s]."
                                              "  Error[%s]." % (self.pool_detail["id"], self.ssid, to_native(error)))

                # Wait for expansion completion unless it is the last request in the candidate list
                if required_expansion_candidate_list:
                    for dummy in range(self.EXPANSION_TIMEOUT_SEC):
                        rc, actions_resp = self.request("storage-systems/%s/storage-pools/%s/action-progress"
                                                        % (self.ssid, self.pool_detail["id"]), ignore_errors=True)
                        if rc == 200:
                            for action in actions_resp:
                                if (action["volumeRef"] in self.storage_pool_volumes and
                                        action["currentAction"] == "remappingDce"):
                                    sleep(1)
                                    estimated_completion_time = action["estimatedTimeToCompletion"]
                                    break
                            else:
                                estimated_completion_time = 0.0
                                break

        return changed_required, estimated_completion_time

    def apply(self):
        """Apply requested state to storage array."""
        changed = False

        if self.state == "present":
            if self.criteria_drive_count is None and self.criteria_min_usable_capacity is None:
                self.module.fail_json(msg="One of criteria_min_usable_capacity or criteria_drive_count must be"
                                          " specified.")
            if self.criteria_drive_count and not self.is_drive_count_valid(self.criteria_drive_count):
                self.module.fail_json(msg="criteria_drive_count must be valid for the specified raid level.")

        self.pool_detail = self.storage_pool
        self.module.log(pformat(self.pool_detail))

        if self.state == "present" and self.erase_secured_drives:
            self.erase_all_available_secured_drives(check_mode=True)

        # Determine whether changes need to be applied to the storage array
        if self.pool_detail:

            if self.state == "absent":
                changed = True

            elif self.state == "present":

                if self.criteria_drive_count and self.criteria_drive_count < len(self.storage_pool_drives):
                    self.module.fail_json(msg="Failed to reduce the size of the storage pool. Array [%s]. Pool [%s]."
                                              % (self.ssid, self.pool_detail["id"]))

                if self.criteria_drive_type and self.criteria_drive_type != self.pool_detail["driveMediaType"]:
                    self.module.fail_json(msg="Failed! It is not possible to modify storage pool media type."
                                              " Array [%s]. Pool [%s]." % (self.ssid, self.pool_detail["id"]))

                if (self.criteria_drive_require_da is not None and self.criteria_drive_require_da !=
                        self.pool_detail["protectionInformationCapabilities"]["protectionInformationCapable"]):
                    self.module.fail_json(msg="Failed! It is not possible to modify DA-capability. Array [%s]."
                                              " Pool [%s]." % (self.ssid, self.pool_detail["id"]))

                # Evaluate current storage pool for required change.
                needs_expansion, estimated_completion_time = self.expand_storage_pool(check_mode=True)
                if needs_expansion:
                    changed = True
                if self.migrate_raid_level(check_mode=True):
                    changed = True
                if self.secure_storage_pool(check_mode=True):
                    changed = True
                if self.set_reserve_drive_count(check_mode=True):
                    changed = True

        elif self.state == "present":
            changed = True

        # Apply changes to storage array
        msg = "No changes were required for the storage pool [%s]."
        if changed and not self.module.check_mode:
            if self.state == "present":
                if self.erase_secured_drives:
                    self.erase_all_available_secured_drives()

                if self.pool_detail:
                    change_list = list()

                    # Expansion needs to occur before raid level migration to account for any sizing needs.
                    expanded, estimated_completion_time = self.expand_storage_pool()
                    if expanded:
                        change_list.append("expanded")
                    if self.migrate_raid_level():
                        change_list.append("raid migration")
                    if self.secure_storage_pool():
                        change_list.append("secured")
                    if self.set_reserve_drive_count():
                        change_list.append("adjusted reserve drive count")

                    if change_list:
                        msg = "Following changes have been applied to the storage pool [%s]: " + ", ".join(change_list)

                    if expanded:
                        msg += "\nThe expansion operation will complete in an estimated %s minutes."\
                               % estimated_completion_time
                else:
                    self.create_storage_pool()
                    msg = "Storage pool [%s] was created."

                    if self.secure_storage_pool():
                        msg = "Storage pool [%s] was created and secured."
                    if self.set_reserve_drive_count():
                        msg += " Adjusted reserve drive count."

            elif self.pool_detail:
                self.delete_storage_pool()
                msg = "Storage pool [%s] removed."

        self.pool_detail = self.storage_pool
        self.module.log(pformat(self.pool_detail))
        self.module.log(msg % self.name)
        self.module.exit_json(msg=msg % self.name, changed=changed, **self.pool_detail)


def main():
    storage_pool = NetAppESeriesStoragePool()
    storage_pool.apply()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Andrew Klychkov (@Andersson007) <aaklychkov@mail.ru>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'supported_by': 'community',
    'status': ['preview']
}

DOCUMENTATION = r'''
---
module: postgresql_membership
short_description: Add or remove PostgreSQL roles from groups
description:
- Adds or removes PostgreSQL roles from groups (other roles).
- Users are roles with login privilege.
- Groups are PostgreSQL roles usually without LOGIN privilege.
- "Common use case:"
- 1) add a new group (groups) by M(postgresql_user) module with I(role_attr_flags=NOLOGIN)
- 2) grant them desired privileges by M(postgresql_privs) module
- 3) add desired PostgreSQL users to the new group (groups) by this module
version_added: '2.8'
options:
  groups:
    description:
    - The list of groups (roles) that need to be granted to or revoked from I(target_roles).
    required: yes
    type: list
    elements: str
    aliases:
    - group
    - source_role
    - source_roles
  target_roles:
    description:
    - The list of target roles (groups will be granted to them).
    required: yes
    type: list
    elements: str
    aliases:
    - target_role
    - users
    - user
  fail_on_role:
    description:
      - If C(yes), fail when group or target_role doesn't exist. If C(no), just warn and continue.
    default: yes
    type: bool
  state:
    description:
    - Membership state.
    - I(state=present) implies the I(groups)must be granted to I(target_roles).
    - I(state=absent) implies the I(groups) must be revoked from I(target_roles).
    type: str
    default: present
    choices: [ absent, present ]
  db:
    description:
    - Name of database to connect to.
    type: str
    aliases:
    - login_db
  session_role:
    description:
    - Switch to session_role after connecting.
      The specified session_role must be a role that the current login_user is a member of.
    - Permissions checking for SQL commands is carried out as though
      the session_role were the one that had logged in originally.
    type: str
seealso:
- module: postgresql_user
- module: postgresql_privs
- module: postgresql_owner
- name: PostgreSQL role membership reference
  description: Complete reference of the PostgreSQL role membership documentation.
  link: https://www.postgresql.org/docs/current/role-membership.html
- name: PostgreSQL role attributes reference
  description: Complete reference of the PostgreSQL role attributes documentation.
  link: https://www.postgresql.org/docs/current/role-attributes.html
author:
- Andrew Klychkov (@Andersson007)
extends_documentation_fragment: postgres
'''

EXAMPLES = r'''
- name: Grant role read_only to alice and bob
  postgresql_membership:
    group: read_only
    target_roles:
    - alice
    - bob
    state: present

# you can also use target_roles: alice,bob,etc to pass the role list

- name: Revoke role read_only and exec_func from bob. Ignore if roles don't exist
  postgresql_membership:
    groups:
    - read_only
    - exec_func
    target_role: bob
    fail_on_role: no
    state: absent
'''

RETURN = r'''
queries:
    description: List of executed queries.
    returned: always
    type: str
    sample: [ "GRANT \"user_ro\" TO \"alice\"" ]
granted:
    description: Dict of granted groups and roles.
    returned: if I(state=present)
    type: dict
    sample: { "ro_group": [ "alice", "bob" ] }
revoked:
    description: Dict of revoked groups and roles.
    returned: if I(state=absent)
    type: dict
    sample: { "ro_group": [ "alice", "bob" ] }
state:
    description: Membership state that tried to be set.
    returned: always
    type: str
    sample: "present"
'''

try:
    from psycopg2.extras import DictCursor
except ImportError:
    # psycopg2 is checked by connect_to_db()
    # from ansible.module_utils.postgres
    pass

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.database import pg_quote_identifier
from ansible.module_utils.postgres import (
    connect_to_db,
    exec_sql,
    get_conn_params,
    postgres_common_argument_spec,
)


class PgMembership(object):
    def __init__(self, module, cursor, groups, target_roles, fail_on_role):
        self.module = module
        self.cursor = cursor
        self.target_roles = [r.strip() for r in target_roles]
        self.groups = [r.strip() for r in groups]
        self.executed_queries = []
        self.granted = {}
        self.revoked = {}
        self.fail_on_role = fail_on_role
        self.non_existent_roles = []
        self.changed = False
        self.__check_roles_exist()

    def grant(self):
        for group in self.groups:
            self.granted[group] = []

            for role in self.target_roles:
                # If role is in a group now, pass:
                if self.__check_membership(group, role):
                    continue

                query = "GRANT %s TO %s" % ((pg_quote_identifier(group, 'role'),
                                            (pg_quote_identifier(role, 'role'))))
                self.changed = exec_sql(self, query, ddl=True)

                if self.changed:
                    self.granted[group].append(role)

        return self.changed

    def revoke(self):
        for group in self.groups:
            self.revoked[group] = []

            for role in self.target_roles:
                # If role is not in a group now, pass:
                if not self.__check_membership(group, role):
                    continue

                query = "REVOKE %s FROM %s" % ((pg_quote_identifier(group, 'role'),
                                               (pg_quote_identifier(role, 'role'))))
                self.changed = exec_sql(self, query, ddl=True)

                if self.changed:
                    self.revoked[group].append(role)

        return self.changed

    def __check_membership(self, src_role, dst_role):
        query = ("SELECT ARRAY(SELECT b.rolname FROM "
                 "pg_catalog.pg_auth_members m "
                 "JOIN pg_catalog.pg_roles b ON (m.roleid = b.oid) "
                 "WHERE m.member = r.oid) "
                 "FROM pg_catalog.pg_roles r "
                 "WHERE r.rolname = '%s'" % dst_role)

        res = exec_sql(self, query, add_to_executed=False)
        membership = []
        if res:
            membership = res[0][0]

        if not membership:
            return False

        if src_role in membership:
            return True

        return False

    def __check_roles_exist(self):
        for group in self.groups:
            if not self.__role_exists(group):
                if self.fail_on_role:
                    self.module.fail_json(msg="Role %s does not exist" % group)
                else:
                    self.module.warn("Role %s does not exist, pass" % group)
                    self.non_existent_roles.append(group)

        for role in self.target_roles:
            if not self.__role_exists(role):
                if self.fail_on_role:
                    self.module.fail_json(msg="Role %s does not exist" % role)
                else:
                    self.module.warn("Role %s does not exist, pass" % role)

                if role not in self.groups:
                    self.non_existent_roles.append(role)

                else:
                    if self.fail_on_role:
                        self.module.exit_json(msg="Role role '%s' is a member of role '%s'" % (role, role))
                    else:
                        self.module.warn("Role role '%s' is a member of role '%s', pass" % (role, role))

        # Update role lists, excluding non existent roles:
        self.groups = [g for g in self.groups if g not in self.non_existent_roles]

        self.target_roles = [r for r in self.target_roles if r not in self.non_existent_roles]

    def __role_exists(self, role):
        return exec_sql(self, "SELECT 1 FROM pg_roles WHERE rolname = '%s'" % role, add_to_executed=False)


# ===========================================
# Module execution.
#


def main():
    argument_spec = postgres_common_argument_spec()
    argument_spec.update(
        groups=dict(type='list', aliases=['group', 'source_role', 'source_roles']),
        target_roles=dict(type='list', aliases=['target_role', 'user', 'users']),
        fail_on_role=dict(type='bool', default=True),
        state=dict(type='str', default='present', choices=['absent', 'present']),
        db=dict(type='str', aliases=['login_db']),
        session_role=dict(type='str'),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    groups = module.params['groups']
    target_roles = module.params['target_roles']
    fail_on_role = module.params['fail_on_role']
    state = module.params['state']

    conn_params = get_conn_params(module, module.params, warn_db_default=False)
    db_connection = connect_to_db(module, conn_params, autocommit=False)
    cursor = db_connection.cursor(cursor_factory=DictCursor)

    ##############
    # Create the object and do main job:

    pg_membership = PgMembership(module, cursor, groups, target_roles, fail_on_role)

    if state == 'present':
        pg_membership.grant()

    elif state == 'absent':
        pg_membership.revoke()

    # Rollback if it's possible and check_mode:
    if module.check_mode:
        db_connection.rollback()
    else:
        db_connection.commit()

    cursor.close()
    db_connection.close()

    # Make return values:
    return_dict = dict(
        changed=pg_membership.changed,
        state=state,
        groups=pg_membership.groups,
        target_roles=pg_membership.target_roles,
        queries=pg_membership.executed_queries,
    )

    if state == 'present':
        return_dict['granted'] = pg_membership.granted
    elif state == 'absent':
        return_dict['revoked'] = pg_membership.revoked

    module.exit_json(**return_dict)


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.1",
    "status": ["preview"],
    "supported_by": "community"
}

DOCUMENTATION = '''
---
module: fmgr_fwobj_address
version_added: "2.8"
notes:
    - Full Documentation at U(https://ftnt-ansible-docs.readthedocs.io/en/latest/).
author:
    - Luke Weighall (@lweighall)
    - Andrew Welsh (@Ghilli3)
    - Jim Huber (@p4r4n0y1ng)
short_description: Allows the management of firewall objects in FortiManager
description:
  -  Allows for the management of IPv4, IPv6, and multicast address objects within FortiManager.

options:
  adom:
    description:
      - The ADOM the configuration should belong to.
    required: false
    default: root

  allow_routing:
    description:
      - Enable/disable use of this address in the static route configuration.
    choices: ['enable', 'disable']
    default: 'disable'

  associated_interface:
    description:
      - Associated interface name.

  cache_ttl:
    description:
      - Minimal TTL of individual IP addresses in FQDN cache. Only applies when type = wildcard-fqdn.

  color:
    description:
      - Color of the object in FortiManager GUI.
      - Takes integers 1-32
    default: 22

  comment:
    description:
      - Comment for the object in FortiManager.

  country:
    description:
      - Country name. Required if type = geographic.

  end_ip:
    description:
      - End IP. Only used when ipv4 = iprange.

  group_members:
    description:
      - Address group member. If this is defined w/out group_name, the operation will fail.

  group_name:
    description:
      - Address group name. If this is defined in playbook task, all other options are ignored.

  ipv4:
    description:
      - Type of IPv4 Object.
      - Must not be specified with either multicast or IPv6 parameters.
    choices: ['ipmask', 'iprange', 'fqdn', 'wildcard', 'geography', 'wildcard-fqdn', 'group']

  ipv4addr:
    description:
      - IP and network mask. If only defining one IP use this parameter. (i.e. 10.7.220.30/255.255.255.255)
      - Can also define subnets (i.e. 10.7.220.0/255.255.255.0)
      - Also accepts CIDR (i.e. 10.7.220.0/24)
      - If Netmask is omitted after IP address, /32 is assumed.
      - When multicast is set to Broadcast Subnet the ipv4addr parameter is used to specify the subnet.

  ipv6:
    description:
      - Puts module into IPv6 mode.
      - Must not be specified with either ipv4 or multicast parameters.
    choices: ['ip', 'iprange', 'group']

  ipv6addr:
    description:
      - IPv6 address in full. (i.e. 2001:0db8:85a3:0000:0000:8a2e:0370:7334)

  fqdn:
    description:
      - Fully qualified domain name.

  mode:
    description:
      - Sets one of three modes for managing the object.
    choices: ['add', 'set', 'delete']
    default: add

  multicast:
    description:
      - Manages Multicast Address Objects.
      - Sets either a Multicast IP Range or a Broadcast Subnet.
      - Must not be specified with either ipv4 or ipv6 parameters.
      - When set to Broadcast Subnet the ipv4addr parameter is used to specify the subnet.
      - Can create IPv4 Multicast Objects (multicastrange and broadcastmask options -- uses start/end-ip and ipv4addr).
    choices: ['multicastrange', 'broadcastmask', 'ip6']

  name:
    description:
      - Friendly Name Address object name in FortiManager.

  obj_id:
    description:
      - Object ID for NSX.

  start_ip:
    description:
      - Start IP. Only used when ipv4 = iprange.

  visibility:
    description:
      - Enable/disable address visibility.
    choices: ['enable', 'disable']
    default: 'enable'

  wildcard:
    description:
      - IP address and wildcard netmask. Required if ipv4 = wildcard.

  wildcard_fqdn:
    description:
      - Wildcard FQDN. Required if ipv4 = wildcard-fqdn.
'''

EXAMPLES = '''
- name: ADD IPv4 IP ADDRESS OBJECT
  fmgr_fwobj_address:
    ipv4: "ipmask"
    ipv4addr: "10.7.220.30/32"
    name: "ansible_v4Obj"
    comment: "Created by Ansible"
    color: "6"

- name: ADD IPv4 IP ADDRESS OBJECT MORE OPTIONS
  fmgr_fwobj_address:
    ipv4: "ipmask"
    ipv4addr: "10.7.220.34/32"
    name: "ansible_v4Obj_MORE"
    comment: "Created by Ansible"
    color: "6"
    allow_routing: "enable"
    cache_ttl: "180"
    associated_interface: "port1"
    obj_id: "123"

- name: ADD IPv4 IP ADDRESS SUBNET OBJECT
  fmgr_fwobj_address:
    ipv4: "ipmask"
    ipv4addr: "10.7.220.0/255.255.255.128"
    name: "ansible_subnet"
    comment: "Created by Ansible"
    mode: "set"

- name: ADD IPv4 IP ADDRESS RANGE OBJECT
  fmgr_fwobj_address:
    ipv4: "iprange"
    start_ip: "10.7.220.1"
    end_ip: "10.7.220.125"
    name: "ansible_range"
    comment: "Created by Ansible"

- name: ADD IPv4 IP ADDRESS WILDCARD OBJECT
  fmgr_fwobj_address:
    ipv4: "wildcard"
    wildcard: "10.7.220.30/255.255.255.255"
    name: "ansible_wildcard"
    comment: "Created by Ansible"

- name: ADD IPv4 IP ADDRESS WILDCARD FQDN OBJECT
  fmgr_fwobj_address:
    ipv4: "wildcard-fqdn"
    wildcard_fqdn: "*.myds.com"
    name: "Synology myds DDNS service"
    comment: "Created by Ansible"

- name: ADD IPv4 IP ADDRESS FQDN OBJECT
  fmgr_fwobj_address:
    ipv4: "fqdn"
    fqdn: "ansible.com"
    name: "ansible_fqdn"
    comment: "Created by Ansible"

- name: ADD IPv4 IP ADDRESS GEO OBJECT
  fmgr_fwobj_address:
    ipv4: "geography"
    country: "usa"
    name: "ansible_geo"
    comment: "Created by Ansible"

- name: ADD IPv6 ADDRESS
  fmgr_fwobj_address:
    ipv6: "ip"
    ipv6addr: "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
    name: "ansible_v6Obj"
    comment: "Created by Ansible"

- name: ADD IPv6 ADDRESS RANGE
  fmgr_fwobj_address:
    ipv6: "iprange"
    start_ip: "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
    end_ip: "2001:0db8:85a3:0000:0000:8a2e:0370:7446"
    name: "ansible_v6range"
    comment: "Created by Ansible"

- name: ADD IPv4 IP ADDRESS GROUP
  fmgr_fwobj_address:
    ipv4: "group"
    group_name: "ansibleIPv4Group"
    group_members: "ansible_fqdn, ansible_wildcard, ansible_range"

- name: ADD IPv6 IP ADDRESS GROUP
  fmgr_fwobj_address:
    ipv6: "group"
    group_name: "ansibleIPv6Group"
    group_members: "ansible_v6Obj, ansible_v6range"

- name: ADD MULTICAST RANGE
  fmgr_fwobj_address:
    multicast: "multicastrange"
    start_ip: "224.0.0.251"
    end_ip: "224.0.0.251"
    name: "ansible_multicastrange"
    comment: "Created by Ansible"

- name: ADD BROADCAST SUBNET
  fmgr_fwobj_address:
    multicast: "broadcastmask"
    ipv4addr: "10.7.220.0/24"
    name: "ansible_broadcastSubnet"
    comment: "Created by Ansible"
'''

RETURN = """
api_result:
  description: full API response, includes status code and message
  returned: always
  type: str
"""


import re
from ansible.module_utils.basic import AnsibleModule, env_fallback
from ansible.module_utils.connection import Connection
from ansible.module_utils.network.fortimanager.fortimanager import FortiManagerHandler
from ansible.module_utils.network.fortimanager.common import FMGBaseException
from ansible.module_utils.network.fortimanager.common import FMGRCommon
from ansible.module_utils.network.fortimanager.common import DEFAULT_RESULT_OBJ
from ansible.module_utils.network.fortimanager.common import FAIL_SOCKET_MSG


def fmgr_fwobj_ipv4(fmgr, paramgram):
    """
    :param fmgr: The fmgr object instance from fortimanager.py
    :type fmgr: class object
    :param paramgram: The formatted dictionary of options to process
    :type paramgram: dict
    :return: The response from the FortiManager
    :rtype: dict
    """
    # EVAL THE MODE PARAMETER FOR SET OR ADD
    if paramgram["mode"] in ['set', 'add']:
        # CREATE THE DATAGRAM DICTIONARY
        # ENSURE THE DATAGRAM KEYS MATCH THE JSON API GUIDE ATTRIBUTES, NOT WHAT IS IN ANSIBLE
        # SOME PARAMETERS SHOWN IN THIS DICTIONARY WE DON'T EVEN ASK THE USER FOR IN PLAYBOOKS BUT ARE REQUIRED
        datagram = {
            "comment": paramgram["comment"],
            "associated-interface": paramgram["associated-interface"],
            "cache-ttl": paramgram["cache-ttl"],
            "name": paramgram["name"],
            "allow-routing": paramgram["allow-routing"],
            "color": paramgram["color"],
            "meta fields": {},
            "dynamic_mapping": [],
            "visibility": paramgram["allow-routing"],
            "type": paramgram["ipv4"],
        }

        # SET THE CORRECT URL BASED ON THE TYPE (WE'RE DOING GROUPS IN THIS METHOD, TOO)
        if datagram["type"] == "group":
            url = '/pm/config/adom/{adom}/obj/firewall/addrgrp'.format(adom=paramgram["adom"])
        else:
            url = '/pm/config/adom/{adom}/obj/firewall/address'.format(adom=paramgram["adom"])

        #########################
        # IF type = 'ipmask'
        #########################
        if datagram["type"] == "ipmask":
            # CREATE THE SUBNET LIST OBJECT
            subnet = []
            # EVAL THE IPV4ADDR INPUT AND SPLIT THE IP ADDRESS FROM THE MASK AND APPEND THEM TO THE SUBNET LIST
            for subnets in paramgram["ipv4addr"].split("/"):
                subnet.append(subnets)

            # CHECK THAT THE SECOND ENTRY IN THE SUBNET LIST (WHAT WAS TO THE RIGHT OF THE / CHARACTER)
            # IS IN SUBNET MASK FORMAT AND NOT CIDR FORMAT.
            # IF IT IS IN CIDR FORMAT, WE NEED TO CONVERT IT TO SUBNET BIT MASK FORMAT FOR THE JSON API
            if not re.match(r'\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}', subnet[1]):
                # IF THE SUBNET PARAMETER INPUT DIDN'T LOOK LIKE xxx.xxx.xxx.xxx TO REGEX...
                # ... RUN IT THROUGH THE CIDR_TO_NETMASK() FUNCTION
                mask = fmgr._tools.cidr_to_netmask(subnet[1])
                # AND THEN UPDATE THE SUBNET LIST OBJECT
                subnet[1] = mask

            # INCLUDE THE SUBNET LIST OBJECT IN THE DATAGRAM DICTIONARY TO BE SUBMITTED
            datagram["subnet"] = subnet

        #########################
        # IF type = 'iprange'
        #########################
        if datagram["type"] == "iprange":
            datagram["start-ip"] = paramgram["start-ip"]
            datagram["end-ip"] = paramgram["end-ip"]
            datagram["subnet"] = ["0.0.0.0", "0.0.0.0"]

        #########################
        # IF type = 'geography'
        #########################
        if datagram["type"] == "geography":
            datagram["country"] = paramgram["country"]

        #########################
        # IF type = 'wildcard'
        #########################
        if datagram["type"] == "wildcard":

            subnet = []
            for subnets in paramgram["wildcard"].split("/"):
                subnet.append(subnets)

            if not re.match(r'\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}', subnet[1]):
                mask = fmgr._tools.cidr_to_netmask(subnet[1])
                subnet[1] = mask

            datagram["wildcard"] = subnet

        #########################
        # IF type = 'wildcard-fqdn'
        #########################
        if datagram["type"] == "wildcard-fqdn":
            datagram["wildcard-fqdn"] = paramgram["wildcard-fqdn"]

        #########################
        # IF type = 'fqdn'
        #########################
        if datagram["type"] == "fqdn":
            datagram["fqdn"] = paramgram["fqdn"]

        #########################
        # IF type = 'group'
        #########################
        if datagram["type"] == "group":
            datagram = {
                "comment": paramgram["comment"],
                "name": paramgram["group_name"],
                "color": paramgram["color"],
                "meta fields": {},
                "dynamic_mapping": [],
                "visibility": paramgram["visibility"]
            }

            members = []
            group_members = paramgram["group_members"].replace(" ", "")
            try:
                for member in group_members.split(","):
                    members.append(member)
            except Exception:
                pass

            datagram["member"] = members

    # EVAL THE MODE PARAMETER FOR DELETE
    if paramgram["mode"] == "delete":
        # IF A GROUP, SET THE CORRECT NAME AND URL FOR THE GROUP ENDPOINT
        if paramgram["ipv4"] == "group":
            datagram = {}
            url = '/pm/config/adom/{adom}/obj/firewall/addrgrp/{name}'.format(adom=paramgram["adom"],
                                                                              name=paramgram["group_name"])
        # OTHERWISE WE'RE JUST GOING TO USE THE ADDRESS ENDPOINT
        else:
            datagram = {}
            url = '/pm/config/adom/{adom}/obj/firewall/address/{name}'.format(adom=paramgram["adom"],
                                                                              name=paramgram["name"])

    response = fmgr.process_request(url, datagram, paramgram["mode"])
    return response


def fmgr_fwobj_ipv6(fmgr, paramgram):
    """
    :param fmgr: The fmgr object instance from fortimanager.py
    :type fmgr: class object
    :param paramgram: The formatted dictionary of options to process
    :type paramgram: dict
    :return: The response from the FortiManager
    :rtype: dict
    """
    # EVAL THE MODE PARAMETER FOR SET OR ADD
    if paramgram["mode"] in ['set', 'add']:
        # CREATE THE DATAGRAM DICTIONARY
        # ENSURE THE DATAGRAM KEYS MATCH THE JSON API GUIDE ATTRIBUTES, NOT WHAT IS IN ANSIBLE
        # SOME PARAMETERS SHOWN IN THIS DICTIONARY WE DON'T EVEN ASK THE USER FOR IN PLAYBOOKS BUT ARE REQUIRED
        datagram = {
            "comment": paramgram["comment"],
            "name": paramgram["name"],
            "color": paramgram["color"],
            "dynamic_mapping": [],
            "visibility": paramgram["visibility"],
            "type": paramgram["ipv6"]
        }

        # SET THE CORRECT URL BASED ON THE TYPE (WE'RE DOING GROUPS IN THIS METHOD, TOO)
        if datagram["type"] == "group":
            url = '/pm/config/adom/{adom}/obj/firewall/addrgrp6'.format(adom=paramgram["adom"])
        else:
            url = '/pm/config/adom/{adom}/obj/firewall/address6'.format(adom=paramgram["adom"])

        #########################
        # IF type = 'ip'
        #########################
        if datagram["type"] == "ip":
            datagram["type"] = "ipprefix"
            datagram["ip6"] = paramgram["ipv6addr"]

        #########################
        # IF type = 'iprange'
        #########################
        if datagram["type"] == "iprange":
            datagram["start-ip"] = paramgram["start-ip"]
            datagram["end-ip"] = paramgram["end-ip"]

        #########################
        # IF type = 'group'
        #########################
        if datagram["type"] == "group":
            datagram = None
            datagram = {
                "comment": paramgram["comment"],
                "name": paramgram["group_name"],
                "color": paramgram["color"],
                "visibility": paramgram["visibility"]
            }

            members = []
            group_members = paramgram["group_members"].replace(" ", "")
            try:
                for member in group_members.split(","):
                    members.append(member)
            except Exception:
                pass

            datagram["member"] = members

    # EVAL THE MODE PARAMETER FOR DELETE
    if paramgram["mode"] == "delete":
        # IF A GROUP, SET THE CORRECT NAME AND URL FOR THE GROUP ENDPOINT
        if paramgram["ipv6"] == "group":
            datagram = {}
            url = '/pm/config/adom/{adom}/obj/firewall/addrgrp6/{name}'.format(adom=paramgram["adom"],
                                                                               name=paramgram["group_name"])
        # OTHERWISE WE'RE JUST GOING TO USE THE ADDRESS ENDPOINT
        else:
            datagram = {}
            url = '/pm/config/adom/{adom}/obj/firewall/address6/{name}'.format(adom=paramgram["adom"],
                                                                               name=paramgram["name"])

    response = fmgr.process_request(url, datagram, paramgram["mode"])
    return response


def fmgr_fwobj_multicast(fmgr, paramgram):
    """
    :param fmgr: The fmgr object instance from fortimanager.py
    :type fmgr: class object
    :param paramgram: The formatted dictionary of options to process
    :type paramgram: dict
    :return: The response from the FortiManager
    :rtype: dict
    """
    # EVAL THE MODE PARAMETER FOR SET OR ADD
    if paramgram["mode"] in ['set', 'add']:
        # CREATE THE DATAGRAM DICTIONARY
        # ENSURE THE DATAGRAM KEYS MATCH THE JSON API GUIDE ATTRIBUTES, NOT WHAT IS IN ANSIBLE
        # SOME PARAMETERS SHOWN IN THIS DICTIONARY WE DON'T EVEN ASK THE USER FOR IN PLAYBOOKS BUT ARE REQUIRED
        datagram = {
            "associated-interface": paramgram["associated-interface"],
            "comment": paramgram["comment"],
            "name": paramgram["name"],
            "color": paramgram["color"],
            "type": paramgram["multicast"],
            "visibility": paramgram["visibility"],
        }

        # SET THE CORRECT URL
        url = '/pm/config/adom/{adom}/obj/firewall/multicast-address'.format(adom=paramgram["adom"])

        #########################
        # IF type = 'multicastrange'
        #########################
        if paramgram["multicast"] == "multicastrange":
            datagram["start-ip"] = paramgram["start-ip"]
            datagram["end-ip"] = paramgram["end-ip"]
            datagram["subnet"] = ["0.0.0.0", "0.0.0.0"]

        #########################
        # IF type = 'broadcastmask'
        #########################
        if paramgram["multicast"] == "broadcastmask":
            # EVAL THE IPV4ADDR INPUT AND SPLIT THE IP ADDRESS FROM THE MASK AND APPEND THEM TO THE SUBNET LIST
            subnet = []
            for subnets in paramgram["ipv4addr"].split("/"):
                subnet.append(subnets)
            # CHECK THAT THE SECOND ENTRY IN THE SUBNET LIST (WHAT WAS TO THE RIGHT OF THE / CHARACTER)
            # IS IN SUBNET MASK FORMAT AND NOT CIDR FORMAT.
            # IF IT IS IN CIDR FORMAT, WE NEED TO CONVERT IT TO SUBNET BIT MASK FORMAT FOR THE JSON API
            if not re.match(r'\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}', subnet[1]):
                # IF THE SUBNET PARAMETER INPUT DIDN'T LOOK LIKE 255.255.255.255 TO REGEX...
                # ... RUN IT THROUGH THE fmgr_cidr_to_netmask() FUNCTION
                mask = fmgr._tools.cidr_to_netmask(subnet[1])
                # AND THEN UPDATE THE SUBNET LIST OBJECT
                subnet[1] = mask

            # INCLUDE THE SUBNET LIST OBJECT IN THE DATAGRAM DICTIONARY TO BE SUBMITTED
            datagram["subnet"] = subnet

    # EVAL THE MODE PARAMETER FOR DELETE
    if paramgram["mode"] == "delete":
        datagram = {
            "name": paramgram["name"]
        }
        # SET THE CORRECT URL FOR DELETE
        url = '/pm/config/adom/{adom}/obj/firewall/multicast-address/{name}'.format(adom=paramgram["adom"],
                                                                                    name=paramgram["name"])

    response = fmgr.process_request(url, datagram, paramgram["mode"])
    return response


def main():
    argument_spec = dict(
        adom=dict(required=False, type="str", default="root"),
        mode=dict(choices=["add", "set", "delete"], type="str", default="add"),

        allow_routing=dict(required=False, type="str", choices=['enable', 'disable'], default="disable"),
        associated_interface=dict(required=False, type="str"),
        cache_ttl=dict(required=False, type="str"),
        color=dict(required=False, type="str", default=22),
        comment=dict(required=False, type="str"),
        country=dict(required=False, type="str"),
        fqdn=dict(required=False, type="str"),
        name=dict(required=False, type="str"),
        start_ip=dict(required=False, type="str"),
        end_ip=dict(required=False, type="str"),
        ipv4=dict(required=False, type="str", choices=['ipmask', 'iprange', 'fqdn', 'wildcard',
                                                       'geography', 'wildcard-fqdn', 'group']),
        visibility=dict(required=False, type="str", choices=['enable', 'disable'], default="enable"),
        wildcard=dict(required=False, type="str"),
        wildcard_fqdn=dict(required=False, type="str"),
        ipv6=dict(required=False, type="str", choices=['ip', 'iprange', 'group']),
        group_members=dict(required=False, type="str"),
        group_name=dict(required=False, type="str"),
        ipv4addr=dict(required=False, type="str"),
        ipv6addr=dict(required=False, type="str"),
        multicast=dict(required=False, type="str", choices=['multicastrange', 'broadcastmask', 'ip6']),
        obj_id=dict(required=False, type="str"),

    )

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=False,
                           mutually_exclusive=[
                               ['ipv4', 'ipv6'],
                               ['ipv4', 'multicast'],
                               ['ipv6', 'multicast']
                           ])
    paramgram = {
        "adom": module.params["adom"],
        "allow-routing": module.params["allow_routing"],
        "associated-interface": module.params["associated_interface"],
        "cache-ttl": module.params["cache_ttl"],
        "color": module.params["color"],
        "comment": module.params["comment"],
        "country": module.params["country"],
        "end-ip": module.params["end_ip"],
        "fqdn": module.params["fqdn"],
        "name": module.params["name"],
        "start-ip": module.params["start_ip"],
        "visibility": module.params["visibility"],
        "wildcard": module.params["wildcard"],
        "wildcard-fqdn": module.params["wildcard_fqdn"],
        "ipv6": module.params["ipv6"],
        "ipv4": module.params["ipv4"],
        "group_members": module.params["group_members"],
        "group_name": module.params["group_name"],
        "ipv4addr": module.params["ipv4addr"],
        "ipv6addr": module.params["ipv6addr"],
        "multicast": module.params["multicast"],
        "mode": module.params["mode"],
        "obj-id": module.params["obj_id"],
    }

    module.paramgram = paramgram
    fmgr = None
    if module._socket_path:
        connection = Connection(module._socket_path)
        fmgr = FortiManagerHandler(connection, module)
        fmgr._tools = FMGRCommon()
    else:
        module.fail_json(**FAIL_SOCKET_MSG)

    results = DEFAULT_RESULT_OBJ
    try:
        if paramgram["ipv4"]:
            results = fmgr_fwobj_ipv4(fmgr, paramgram)

        elif paramgram["ipv6"]:
            results = fmgr_fwobj_ipv6(fmgr, paramgram)

        elif paramgram["multicast"]:
            results = fmgr_fwobj_multicast(fmgr, paramgram)

        fmgr.govern_response(module=module, results=results,
                             ansible_facts=fmgr.construct_ansible_facts(results, module.params, paramgram))

    except Exception as err:
        raise FMGBaseException(err)

    if results is not None:
        return module.exit_json(**results[1])
    else:
        return module.exit_json(msg="Couldn't find a proper ipv4 or ipv6 or multicast parameter "
                                    "to run in the logic tree. Exiting...")


if __name__ == "__main__":
    main()



# (c) 2012, Daniel Hokka Zakrisson <daniel@hozac.com>
# (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = """
    lookup: file
    author: Daniel Hokka Zakrisson <daniel@hozac.com>
    version_added: "0.9"
    short_description: read file contents
    description:
        - This lookup returns the contents from a file on the Ansible controller's file system.
    options:
      _terms:
        description: path(s) of files to read
        required: True
      rstrip:
        description: whether or not to remove whitespace from the ending of the looked-up file
        type: bool
        required: False
        default: True
      lstrip:
        description: whether or not to remove whitespace from the beginning of the looked-up file
        type: bool
        required: False
        default: False
    notes:
      - if read in variable context, the file can be interpreted as YAML if the content is valid to the parser.
      - this lookup does not understand 'globing', use the fileglob lookup instead.
"""

EXAMPLES = """
- debug: msg="the value of foo.txt is {{lookup('file', '/etc/foo.txt') }}"

- name: display multiple file contents
  debug: var=item
  with_file:
    - "/path/to/foo.txt"
    - "bar.txt"  # will be looked in files/ dir relative to play or in role
    - "/path/to/biz.txt"
"""

RETURN = """
  _raw:
    description:
      - content of file(s)
"""

from ansible.errors import AnsibleError, AnsibleParserError
from ansible.plugins.lookup import LookupBase
from ansible.module_utils._text import to_text
from ansible.utils.display import Display

display = Display()


class LookupModule(LookupBase):

    def run(self, terms, variables=None, **kwargs):

        ret = []

        for term in terms:
            display.debug("File lookup term: %s" % term)

            # Find the file in the expected search path
            lookupfile = self.find_file_in_search_path(variables, 'files', term)
            display.vvvv(u"File lookup using %s as file" % lookupfile)
            try:
                if lookupfile:
                    b_contents, show_data = self._loader._get_file_contents(lookupfile)
                    contents = to_text(b_contents, errors='surrogate_or_strict')
                    if kwargs.get('lstrip', False):
                        contents = contents.lstrip()
                    if kwargs.get('rstrip', True):
                        contents = contents.rstrip()
                    ret.append(contents)
                else:
                    raise AnsibleParserError()
            except AnsibleParserError:
                raise AnsibleError("could not locate file in lookup: %s" % term)

        return ret



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# (c) 2018, Yanis Guenane <yanis+ansible@guenane.org>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: scaleway_server_info
short_description: Gather information about the Scaleway servers available.
description:
  - Gather information about the Scaleway servers available.
version_added: "2.9"
author:
  - "Yanis Guenane (@Spredzy)"
  - "Remy Leone (@sieben)"
extends_documentation_fragment: scaleway
options:
  region:
    description:
      - Scaleway region to use (for example C(par1)).
    required: true
    choices:
      - ams1
      - EMEA-NL-EVS
      - par1
      - EMEA-FR-PAR1
'''

EXAMPLES = r'''
- name: Gather Scaleway servers information
  scaleway_server_info:
    region: par1
  register: result

- debug:
    msg: "{{ result.scaleway_server_info }}"
'''

RETURN = r'''
---
scaleway_server_info:
  description: Response from Scaleway API
  returned: success
  type: complex
  sample:
    "scaleway_server_info": [
        {
            "arch": "x86_64",
            "boot_type": "local",
            "bootscript": {
                "architecture": "x86_64",
                "bootcmdargs": "LINUX_COMMON scaleway boot=local nbd.max_part=16",
                "default": true,
                "dtb": "",
                "id": "b1e68c26-a19c-4eac-9222-498b22bd7ad9",
                "initrd": "http://169.254.42.24/initrd/initrd-Linux-x86_64-v3.14.5.gz",
                "kernel": "http://169.254.42.24/kernel/x86_64-mainline-lts-4.4-4.4.127-rev1/vmlinuz-4.4.127",
                "organization": "11111111-1111-4111-8111-111111111111",
                "public": true,
                "title": "x86_64 mainline 4.4.127 rev1"
            },
            "commercial_type": "START1-XS",
            "creation_date": "2018-08-14T21:36:56.271545+00:00",
            "dynamic_ip_required": false,
            "enable_ipv6": false,
            "extra_networks": [],
            "hostname": "scw-e0d256",
            "id": "12f19bc7-108c-4517-954c-e6b3d0311363",
            "image": {
                "arch": "x86_64",
                "creation_date": "2018-04-26T12:42:21.619844+00:00",
                "default_bootscript": {
                    "architecture": "x86_64",
                    "bootcmdargs": "LINUX_COMMON scaleway boot=local nbd.max_part=16",
                    "default": true,
                    "dtb": "",
                    "id": "b1e68c26-a19c-4eac-9222-498b22bd7ad9",
                    "initrd": "http://169.254.42.24/initrd/initrd-Linux-x86_64-v3.14.5.gz",
                    "kernel": "http://169.254.42.24/kernel/x86_64-mainline-lts-4.4-4.4.127-rev1/vmlinuz-4.4.127",
                    "organization": "11111111-1111-4111-8111-111111111111",
                    "public": true,
                    "title": "x86_64 mainline 4.4.127 rev1"
                },
                "extra_volumes": [],
                "from_server": null,
                "id": "67375eb1-f14d-4f02-bb42-6119cecbde51",
                "modification_date": "2018-04-26T12:49:07.573004+00:00",
                "name": "Ubuntu Xenial",
                "organization": "51b656e3-4865-41e8-adbc-0c45bdd780db",
                "public": true,
                "root_volume": {
                    "id": "020b8d61-3867-4a0e-84a4-445c5393e05d",
                    "name": "snapshot-87fc282d-f252-4262-adad-86979d9074cf-2018-04-26_12:42",
                    "size": 25000000000,
                    "volume_type": "l_ssd"
                },
                "state": "available"
            },
            "ipv6": null,
            "location": {
                "cluster_id": "5",
                "hypervisor_id": "412",
                "node_id": "2",
                "platform_id": "13",
                "zone_id": "par1"
            },
            "maintenances": [],
            "modification_date": "2018-08-14T21:37:28.630882+00:00",
            "name": "scw-e0d256",
            "organization": "3f709602-5e6c-4619-b80c-e841c89734af",
            "private_ip": "10.14.222.131",
            "protected": false,
            "public_ip": {
                "address": "163.172.170.197",
                "dynamic": false,
                "id": "ea081794-a581-4495-8451-386ddaf0a451"
            },
            "security_group": {
                "id": "a37379d2-d8b0-4668-9cfb-1233fc436f7e",
                "name": "Default security group"
            },
            "state": "running",
            "state_detail": "booted",
            "tags": [],
            "volumes": {
                "0": {
                    "creation_date": "2018-08-14T21:36:56.271545+00:00",
                    "export_uri": "device://dev/vda",
                    "id": "68386fae-4f55-4fbf-aabb-953036a85872",
                    "modification_date": "2018-08-14T21:36:56.271545+00:00",
                    "name": "snapshot-87fc282d-f252-4262-adad-86979d9074cf-2018-04-26_12:42",
                    "organization": "3f709602-5e6c-4619-b80c-e841c89734af",
                    "server": {
                        "id": "12f19bc7-108c-4517-954c-e6b3d0311363",
                        "name": "scw-e0d256"
                    },
                    "size": 25000000000,
                    "state": "available",
                    "volume_type": "l_ssd"
                }
            }
        }
    ]
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.scaleway import (
    Scaleway,
    ScalewayException,
    scaleway_argument_spec,
    SCALEWAY_LOCATION,
)


class ScalewayServerInfo(Scaleway):

    def __init__(self, module):
        super(ScalewayServerInfo, self).__init__(module)
        self.name = 'servers'

        region = module.params["region"]
        self.module.params['api_url'] = SCALEWAY_LOCATION[region]["api_endpoint"]


def main():
    argument_spec = scaleway_argument_spec()
    argument_spec.update(dict(
        region=dict(required=True, choices=SCALEWAY_LOCATION.keys()),
    ))

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    try:
        module.exit_json(
            scaleway_server_info=ScalewayServerInfo(module).get_resources()
        )
    except ScalewayException as exc:
        module.fail_json(msg=exc.message)


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'network'}


DOCUMENTATION = '''
---
module: nxos_aaa_server_host
extends_documentation_fragment: nxos
version_added: "2.2"
short_description: Manages AAA server host-specific configuration.
description:
    - Manages AAA server host-specific configuration.
author: Jason Edelman (@jedelman8)
notes:
    - Tested against NXOSv 7.3.(0)D1(1) on VIRL
    - Changes to the host key (shared secret) are not idempotent for type 0.
    - If C(state=absent) removes the whole host configuration.
options:
    server_type:
        description:
            - The server type is either radius or tacacs.
        required: true
        choices: ['radius', 'tacacs']
    address:
        description:
            - Address or name of the radius or tacacs host.
        required: true
    key:
        description:
            - Shared secret for the specified host or keyword 'default'.
    encrypt_type:
        description:
            - The state of encryption applied to the entered key.
              O for clear text, 7 for encrypted. Type-6 encryption is
              not supported.
        choices: ['0', '7']
    host_timeout:
        description:
            - Timeout period for specified host, in seconds or keyword 'default.
              Range is 1-60.
    auth_port:
        description:
            - Alternate UDP port for RADIUS authentication or keyword 'default'.
    acct_port:
        description:
            - Alternate UDP port for RADIUS accounting or keyword 'default'.
    tacacs_port:
        description:
            - Alternate TCP port TACACS Server or keyword 'default'.
    state:
        description:
            - Manage the state of the resource.
        default: present
        choices: ['present','absent']
'''
EXAMPLES = '''
# Radius Server Host Basic settings
  - name: "Radius Server Host Basic settings"
    nxos_aaa_server_host:
        state: present
        server_type: radius
        address: 1.2.3.4
        acct_port: 2084
        host_timeout: 10

# Radius Server Host Key Configuration
  - name: "Radius Server Host Key Configuration"
    nxos_aaa_server_host:
        state: present
        server_type: radius
        address: 1.2.3.4
        key: hello
        encrypt_type: 7

# TACACS Server Host Configuration
  - name: "Tacacs Server Host Configuration"
    nxos_aaa_server_host:
        state: present
        server_type: tacacs
        tacacs_port: 89
        host_timeout: 10
        address: 5.6.7.8

'''

RETURN = '''
proposed:
    description: k/v pairs of parameters passed into module
    returned: always
    type: dict
    sample: {"address": "1.2.3.4", "auth_port": "2084",
            "host_timeout": "10", "server_type": "radius"}
existing:
    description:
        - k/v pairs of existing configuration
    returned: always
    type: dict
    sample: {}
end_state:
    description: k/v pairs of configuration after module execution
    returned: always
    type: dict
    sample: {"address": "1.2.3.4", "auth_port": "2084",
            "host_timeout": "10", "server_type": "radius"}
updates:
    description: command sent to the device
    returned: always
    type: list
    sample: ["radius-server host 1.2.3.4 auth-port 2084 timeout 10"]
changed:
    description: check to see if a change was made on the device
    returned: always
    type: bool
    sample: true
'''
import re

from ansible.module_utils.network.nxos.nxos import load_config, run_commands
from ansible.module_utils.network.nxos.nxos import get_capabilities, nxos_argument_spec
from ansible.module_utils.basic import AnsibleModule


def execute_show_command(command, module):
    device_info = get_capabilities(module)
    network_api = device_info.get('network_api', 'nxapi')

    if network_api == 'cliconf':
        cmds = [command]
        body = run_commands(module, cmds)
    elif network_api == 'nxapi':
        cmds = {'command': command, 'output': 'text'}
        body = run_commands(module, cmds)

    return body


def flatten_list(command_lists):
    flat_command_list = []
    for command in command_lists:
        if isinstance(command, list):
            flat_command_list.extend(command)
        else:
            flat_command_list.append(command)
    return flat_command_list


def get_aaa_host_info(module, server_type, address):
    aaa_host_info = {}
    command = 'show run | inc {0}-server.host.{1}'.format(server_type, address)

    body = execute_show_command(command, module)[0]
    if body:
        try:
            if 'radius' in body:
                pattern = (r'\S+ host \S+(?:\s+key 7\s+(\S+))?(?:\s+auth-port (\d+))?'
                           r'(?:\s+acct-port (\d+))?(?:\s+authentication)?'
                           r'(?:\s+accounting)?(?:\s+timeout (\d+))?')
                match = re.search(pattern, body)
                aaa_host_info['key'] = match.group(1)
                if aaa_host_info['key']:
                    aaa_host_info['key'] = aaa_host_info['key'].replace('"', '')
                    aaa_host_info['encrypt_type'] = '7'
                aaa_host_info['auth_port'] = match.group(2)
                aaa_host_info['acct_port'] = match.group(3)
                aaa_host_info['host_timeout'] = match.group(4)
            elif 'tacacs' in body:
                pattern = (r'\S+ host \S+(?:\s+key 7\s+(\S+))?(?:\s+port (\d+))?'
                           r'(?:\s+timeout (\d+))?')
                match = re.search(pattern, body)
                aaa_host_info['key'] = match.group(1)
                if aaa_host_info['key']:
                    aaa_host_info['key'] = aaa_host_info['key'].replace('"', '')
                    aaa_host_info['encrypt_type'] = '7'
                aaa_host_info['tacacs_port'] = match.group(2)
                aaa_host_info['host_timeout'] = match.group(3)

            aaa_host_info['server_type'] = server_type
            aaa_host_info['address'] = address
        except TypeError:
            return {}
    else:
        return {}

    return aaa_host_info


def config_aaa_host(server_type, address, params, existing):
    cmds = []
    cmd_str = '{0}-server host {1}'.format(server_type, address)
    cmd_no_str = 'no ' + cmd_str

    key = params.get('key')
    enc_type = params.get('encrypt_type', '')

    defval = False
    nondef = False

    if key:
        if key != 'default':
            cmds.append(cmd_str + ' key {0} {1}'.format(enc_type, key))
        else:
            cmds.append(cmd_no_str + ' key 7 {0}'.format(existing.get('key')))

    locdict = {'auth_port': 'auth-port', 'acct_port': 'acct-port',
               'tacacs_port': 'port', 'host_timeout': 'timeout'}

    # platform CLI needs the keywords in the following order
    for key in ['auth_port', 'acct_port', 'tacacs_port', 'host_timeout']:
        item = params.get(key)
        if item:
            if item != 'default':
                cmd_str += ' {0} {1}'.format(locdict.get(key), item)
                nondef = True
            else:
                cmd_no_str += ' {0} 1'.format(locdict.get(key))
                defval = True
    if defval:
        cmds.append(cmd_no_str)
    if nondef or not existing:
        cmds.append(cmd_str)

    return cmds


def main():
    argument_spec = dict(
        server_type=dict(choices=['radius', 'tacacs'], required=True),
        address=dict(type='str', required=True),
        key=dict(type='str'),
        encrypt_type=dict(type='str', choices=['0', '7']),
        host_timeout=dict(type='str'),
        auth_port=dict(type='str'),
        acct_port=dict(type='str'),
        tacacs_port=dict(type='str'),
        state=dict(choices=['absent', 'present'], default='present'),
    )

    argument_spec.update(nxos_argument_spec)

    module = AnsibleModule(argument_spec=argument_spec,
                           supports_check_mode=True)

    warnings = list()

    server_type = module.params['server_type']
    address = module.params['address']
    key = module.params['key']
    encrypt_type = module.params['encrypt_type']
    host_timeout = module.params['host_timeout']
    auth_port = module.params['auth_port']
    acct_port = module.params['acct_port']
    tacacs_port = module.params['tacacs_port']
    state = module.params['state']

    args = dict(server_type=server_type, address=address, key=key,
                encrypt_type=encrypt_type, host_timeout=host_timeout,
                auth_port=auth_port, acct_port=acct_port,
                tacacs_port=tacacs_port)

    proposed = dict((k, v) for k, v in args.items() if v is not None)
    changed = False

    if encrypt_type and not key:
        module.fail_json(msg='encrypt_type must be used with key')

    if tacacs_port and server_type != 'tacacs':
        module.fail_json(
            msg='tacacs_port can only be used with server_type=tacacs')

    if (auth_port or acct_port) and server_type != 'radius':
        module.fail_json(msg='auth_port and acct_port can only be used'
                             'when server_type=radius')

    existing = get_aaa_host_info(module, server_type, address)
    end_state = existing

    commands = []
    delta = {}
    if state == 'present':
        if not existing:
            delta = proposed
        else:
            for key, value in proposed.items():
                if key == 'encrypt_type':
                    delta[key] = value
                if value != existing.get(key):
                    if value != 'default' or existing.get(key):
                        delta[key] = value

        command = config_aaa_host(server_type, address, delta, existing)
        if command:
            commands.append(command)

    elif state == 'absent':
        intersect = dict(
            set(proposed.items()).intersection(existing.items()))
        if intersect.get('address') and intersect.get('server_type'):
            command = 'no {0}-server host {1}'.format(
                intersect.get('server_type'), intersect.get('address'))
            commands.append(command)

    cmds = flatten_list(commands)
    if cmds:
        if module.check_mode:
            module.exit_json(changed=True, commands=cmds)
        else:
            changed = True
            load_config(module, cmds)
            end_state = get_aaa_host_info(module, server_type, address)

    results = {}
    results['proposed'] = proposed
    results['existing'] = existing
    results['updates'] = cmds
    results['changed'] = changed
    results['warnings'] = warnings
    results['end_state'] = end_state

    module.exit_json(**results)


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright (c) 2019 Yunge Zhu, (@yungezz)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
module: azure_rm_roleassignment_info
version_added: "2.9"
short_description: Gets Azure Role Assignment facts
description:
    - Gets facts of Azure Role Assignment.

options:
    scope:
        description:
            - The scope that the role assignment applies to.
            - For example, use /subscriptions/{subscription-id}/ for a subscription.
            - /subscriptions/{subscription-id}/resourceGroups/{resourcegroup-name} for a resource group.
            - /subscriptions/{subscription-id}/resourceGroups/{resourcegroup-name}/providers/{resource-provider}/{resource-type}/{resource-name} for a resource.
    name:
        description:
            - Name of role assignment.
            - Mutual exclusive with I(assignee).
    assignee:
        description:
            - Object id of a user, group or service principal.
            - Mutually exclusive with I(name).
    role_definition_id:
        description:
            - Resource id of role definition.

extends_documentation_fragment:
    - azure

author:
    - Yunge Zhu(@yungezz)

'''

EXAMPLES = '''
    - name: Get role assignments for specific service principal
      azure_rm_roleassignment_info:
        assignee: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx

    - name: Get role assignments for specific scope
      azure_rm_roleassignment_info:
        scope: /subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
'''

RETURN = '''
roleassignments:
    description:
        - List of role assignments.
    returned: always
    type: complex
    contains:
        id:
            description:
                - Id of role assignment.
            type: str
            returned: always
            sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/providers/Microsoft.Authorization/roleAssignments/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx"
        name:
            description:
                - Name of role assignment.
            type: str
            returned: always
            sample: myRoleAssignment
        type:
            description:
                - Type of role assignment.
            type: str
            returned: always
            sample: custom
        principal_id:
            description:
                - Principal Id of the role assigned to.
            type: str
            returned: always
            sample: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
        role_definition_id:
            description:
                - Role definition id that was assigned to principal_id.
            type: str
            returned: always
            sample: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
        scope:
            description:
                - The role assignment scope.
            type: str
            returned: always
            sample: /subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
'''

import time
from ansible.module_utils.azure_rm_common import AzureRMModuleBase

try:
    from msrestazure.azure_exceptions import CloudError
    from msrest.serialization import Model
    from azure.mgmt.authorization import AuthorizationManagementClient

except ImportError:
    # This is handled in azure_rm_common
    pass


def roleassignment_to_dict(assignment):
    return dict(
        id=assignment.id,
        name=assignment.name,
        type=assignment.type,
        principal_id=assignment.principal_id,
        role_definition_id=assignment.role_definition_id,
        scope=assignment.scope
    )


class AzureRMRoleAssignmentInfo(AzureRMModuleBase):

    def __init__(self):
        self.module_arg_spec = dict(
            name=dict(
                type='str'
            ),
            scope=dict(
                type='str'
            ),
            assignee=dict(
                type='str'
            ),
            role_definition_id=dict(
                type='str'
            )
        )

        self.name = None
        self.scope = None
        self.assignee = None
        self.role_definition_id = None

        self.results = dict(
            changed=False
        )

        self._client = None

        mutually_exclusive = [['name', 'assignee']]

        super(AzureRMRoleAssignmentInfo, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                        supports_tags=False,
                                                        mutually_exclusive=mutually_exclusive)

    def exec_module(self, **kwargs):
        """Main module execution method"""
        is_old_facts = self.module._name == 'azure_rm_roleassignment_facts'
        if is_old_facts:
            self.module.deprecate("The 'azure_rm_roleassignment_facts' module has been renamed to 'azure_rm_roleassignment_info'", version='2.13')

        for key in list(self.module_arg_spec.keys()):
            if hasattr(self, key):
                setattr(self, key, kwargs[key])

        # get management client
        self._client = self.get_mgmt_svc_client(AuthorizationManagementClient,
                                                base_url=self._cloud_environment.endpoints.resource_manager,
                                                api_version="2018-01-01-preview")

        if self.name:
            self.results['roleassignments'] = self.get_by_name()
        elif self.assignee:
            self.results['roleassignments'] = self.get_by_assignee()
        elif self.scope:
            self.results['roleassignments'] = self.list_by_scope()
        else:
            self.fail("Please specify name or assignee")

        return self.results

    def get_by_name(self):
        '''
        Gets the properties of the specified role assignment by name.

        :return: deserialized role assignment dictionary
        '''
        self.log("Gets role assignment {0} by name".format(self.name))

        results = []

        try:
            response = self._client.role_assignments.get(scope=self.scope, role_assignment_name=self.name)

            if response:
                response = roleassignment_to_dict(response)

                if self.role_definition_id:
                    if self.role_definition_id == response['role_definition_id']:
                        results = [response]
                else:
                    results = [response]

        except CloudError as ex:
            self.log("Didn't find role assignment {0} in scope {1}".format(self.name, self.scope))

        return results

    def get_by_assignee(self):
        '''
        Gets the role assignments by assignee.

        :return: deserialized role assignment dictionary
        '''
        self.log("Gets role assignment {0} by name".format(self.name))

        results = []
        filter = "principalId eq '{0}'".format(self.assignee)
        try:
            response = list(self._client.role_assignments.list(filter=filter))

            if response and len(response) > 0:
                response = [roleassignment_to_dict(a) for a in response]

                if self.role_definition_id:
                    for r in response:
                        if r['role_definition_id'] == self.role_definition_id:
                            results.append(r)
                else:
                    results = response

        except CloudError as ex:
            self.log("Didn't find role assignments to assignee {0}".format(self.assignee))

        return results

    def list_by_scope(self):
        '''
        Lists the role assignments by specific scope.

        :return: deserialized role assignment dictionary
        '''
        self.log("Lists role assignment by scope {0}".format(self.scope))

        results = []
        try:
            response = list(self._client.role_assignments.list_for_scope(scope=self.scope, filter='atScope()'))

            if response and len(response) > 0:
                response = [roleassignment_to_dict(a) for a in response]

                if self.role_definition_id:
                    for r in response:
                        if r['role_definition_id'] == self.role_definition_id:
                            results.append(r)
                else:
                    results = response

        except CloudError as ex:
            self.log("Didn't find role assignments to scope {0}".format(self.scope))

        return results


def main():
    """Main execution"""
    AzureRMRoleAssignmentInfo()


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'network'}


DOCUMENTATION = '''
---
module: nxos_rpm
extends_documentation_fragment: nxos
version_added: "2.7"
short_description: Install patch or feature rpms on Cisco NX-OS devices.
description:
    - Install software maintenance upgrade (smu) RPMS and
      3rd party RPMS on Cisco NX-OS devices.
author: Sai Chintalapudi (@saichint)
notes:
    - Tested against NXOSv 7.0(3)I2(5), 7.0(3)I4(6), 7.0(3)I5(3),
      7.0(3)I6(1), 7.0(3)I7(3)
    - For patches, the minimum platform version needed is 7.0(3)I2(5)
    - For feature rpms, the minimum platform version needed is 7.0(3)I6(1)
    - The module manages the entire RPM lifecycle (Add, activate, commit, deactivate, remove)
    - For reload patches, this module is NOT idempotent until the patch is
      committed.
options:
    pkg:
        description:
            - Name of the RPM package.
        required: true
    file_system:
        description:
            - The remote file system of the device. If omitted,
              devices that support a file_system parameter will use
              their default values.
        default: bootflash
    aggregate:
        description:
            - List of RPM/patch definitions.
    state:
        description:
            - If the state is present, the rpm will be installed,
              If the state is absent, it will be removed.
        default: present
        choices: ['present', 'absent']
'''

EXAMPLES = '''
- nxos_rpm:
    pkg: "nxos.sample-n9k_ALL-1.0.0-7.0.3.I7.3.lib32_n9000.rpm"
'''

RETURN = '''
commands:
    description: commands sent to the device
    returned: always
    type: list
    sample: ["install add bootflash:nxos.sample-n9k_ALL-1.0.0-7.0.3.I7.3.lib32_n9000.rpm forced",
             "install activate nxos.sample-n9k_ALL-1.0.0-7.0.3.I7.3.lib32_n9000 forced",
             "install commit nxos.sample-n9k_ALL-1.0.0-7.0.3.I7.3.lib32_n9000"]
'''


import time

from copy import deepcopy

from ansible.module_utils.network.nxos.nxos import load_config, run_commands
from ansible.module_utils.network.nxos.nxos import nxos_argument_spec, check_args
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.network.common.utils import remove_default_spec


def execute_show_command(command, module):
    iteration = 0
    cmds = [{
        'command': command,
        'output': 'text',
    }]

    while iteration < 10:
        body = run_commands(module, cmds)[0]
        if body:
            return body
        else:
            time.sleep(2)
            iteration += 1


def remote_file_exists(module, dst, file_system):
    command = 'dir {0}:/{1}'.format(file_system, dst)
    body = execute_show_command(command, module)
    if 'No such file' in body:
        return False
    return True


def config_cmd_operation(module, cmd):
    iteration = 0
    while iteration < 10:
        msg = load_config(module, [cmd], True)
        if msg:
            if 'another install operation is in progress' in msg[0].lower() or 'failed' in msg[0].lower():
                time.sleep(2)
                iteration += 1
            else:
                return
        else:
            return


def validate_operation(module, show_cmd, cfg_cmd, pkg, pkg_not_present):
    iteration = 0
    while iteration < 10:
        body = execute_show_command(show_cmd, module)
        if pkg_not_present:
            if pkg not in body:
                return
        else:
            if pkg in body:
                return
        time.sleep(2)
        iteration += 1

    err = 'Operation "{0}" Failed'.format(cfg_cmd)
    module.fail_json(msg=err)


def add_operation(module, show_cmd, file_system, full_pkg, pkg):
    cmd = 'install add {0}:{1}'.format(file_system, full_pkg)
    config_cmd_operation(module, cmd)
    validate_operation(module, show_cmd, cmd, pkg, False)
    return cmd


def activate_operation(module, show_cmd, pkg):
    cmd = 'install activate {0} forced'.format(pkg)
    config_cmd_operation(module, cmd)
    validate_operation(module, show_cmd, cmd, pkg, False)
    return cmd


def activate_reload(module, pkg, flag):
    iteration = 0
    if flag:
        cmd = 'install activate {0} forced'.format(pkg)
    else:
        cmd = 'install deactivate {0} forced'.format(pkg)
    opts = {'ignore_timeout': True}
    while iteration < 10:
        msg = load_config(module, [cmd], True, opts)
        if msg:
            if isinstance(msg[0], int):
                if msg[0] == -32603:
                    return cmd
            elif isinstance(msg[0], str):
                if 'another install operation is in progress' in msg[0].lower() or 'failed' in msg[0].lower():
                    time.sleep(2)
                    iteration += 1


def commit_operation(module, show_cmd, pkg):
    cmd = 'install commit {0}'.format(pkg)
    config_cmd_operation(module, cmd)
    validate_operation(module, show_cmd, cmd, pkg, False)
    return cmd


def deactivate_operation(module, show_cmd, pkg, flag):
    cmd = 'install deactivate {0} forced'.format(pkg)
    config_cmd_operation(module, cmd)
    validate_operation(module, show_cmd, cmd, pkg, flag)
    return cmd


def remove_operation(module, show_cmd, pkg):
    cmd = 'install remove {0} forced'.format(pkg)
    config_cmd_operation(module, cmd)
    validate_operation(module, show_cmd, cmd, pkg, True)
    return cmd


def install_remove_rpm(module, full_pkg, file_system, state):
    commands = []
    reload_patch = False

    splitted_pkg = full_pkg.split('.')
    pkg = '.'.join(splitted_pkg[0:-1])

    show_inactive = 'show install inactive'
    show_active = 'show install active'
    show_commit = 'show install committed'
    show_patches = 'show install patches'
    show_pkg_info = 'show install pkg-info {0}'.format(pkg)

    if state == 'present':
        inactive_body = execute_show_command(show_inactive, module)
        active_body = execute_show_command(show_active, module)

        if pkg not in inactive_body and pkg not in active_body:
            commands.append(add_operation(module, show_inactive, file_system, full_pkg, pkg))

        patch_type_body = execute_show_command(show_pkg_info, module)
        if patch_type_body and 'Patch Type    :  reload' in patch_type_body:
            # This is reload smu/patch rpm
            reload_patch = True

        if pkg not in active_body:
            if reload_patch:
                commands.append(activate_reload(module, pkg, True))
                return commands
            else:
                commands.append(activate_operation(module, show_active, pkg))

        commit_body = execute_show_command(show_commit, module)
        if pkg not in commit_body:
            patch_body = execute_show_command(show_patches, module)
            if pkg in patch_body:
                # This is smu/patch rpm
                commands.append(commit_operation(module, show_active, pkg))
            else:
                err = 'Operation "install activate {0} forced" Failed'.format(pkg)
                module.fail_json(msg=err)

    else:
        commit_body = execute_show_command(show_commit, module)
        active_body = execute_show_command(show_active, module)

        patch_type_body = execute_show_command(show_pkg_info, module)
        if patch_type_body and 'Patch Type    :  reload' in patch_type_body:
            # This is reload smu/patch rpm
            reload_patch = True

        if pkg in commit_body and pkg in active_body:
            if reload_patch:
                commands.append(activate_reload(module, pkg, False))
                return commands
            else:
                commands.append(deactivate_operation(module, show_active, pkg, True))
                commit_body = execute_show_command(show_commit, module)
                if pkg in commit_body:
                    # This is smu/patch rpm
                    commands.append(commit_operation(module, show_inactive, pkg))
                commands.append(remove_operation(module, show_inactive, pkg))

        elif pkg in commit_body:
            # This is smu/patch rpm
            commands.append(commit_operation(module, show_inactive, pkg))
            commands.append(remove_operation(module, show_inactive, pkg))

        elif pkg in active_body:
            # This is smu/patch rpm
            if reload_patch:
                commands.append(activate_reload(module, pkg, False))
                return commands
            else:
                commands.append(deactivate_operation(module, show_inactive, pkg, False))
                commands.append(remove_operation(module, show_inactive, pkg))

        else:
            inactive_body = execute_show_command(show_inactive, module)
            if pkg in inactive_body:
                commands.append(remove_operation(module, show_inactive, pkg))

    return commands


def main():
    element_spec = dict(
        pkg=dict(type='str'),
        file_system=dict(type='str', default='bootflash'),
        state=dict(choices=['absent', 'present'], default='present')
    )

    aggregate_spec = deepcopy(element_spec)
    aggregate_spec['pkg'] = dict(required=True)

    # remove default in aggregate spec, to handle common arguments
    remove_default_spec(aggregate_spec)

    argument_spec = dict(
        aggregate=dict(type='list', elements='dict', options=aggregate_spec)
    )

    argument_spec.update(element_spec)
    argument_spec.update(nxos_argument_spec)

    required_one_of = [['pkg', 'aggregate']]
    mutually_exclusive = [['pkg', 'aggregate']]

    module = AnsibleModule(argument_spec=argument_spec,
                           required_one_of=required_one_of,
                           mutually_exclusive=mutually_exclusive,
                           supports_check_mode=False)

    warnings = list()
    results = {'changed': False, 'commands': [], 'warnings': warnings}

    aggregate = module.params.get('aggregate')
    objects = []
    if aggregate:
        for item in aggregate:
            for key in item:
                if item.get(key) is None:
                    item[key] = module.params[key]

            d = item.copy()
            objects.append(d)
    else:
        objects.append({
            'pkg': module.params['pkg'],
            'file_system': module.params['file_system'],
            'state': module.params['state']
        })

    for obj in objects:
        if obj['state'] == 'present':
            remote_exists = remote_file_exists(module, obj['pkg'], file_system=obj['file_system'])

            if not remote_exists:
                module.fail_json(
                    msg="The requested package doesn't exist on the device"
                )

        cmds = install_remove_rpm(module, obj['pkg'], obj['file_system'], obj['state'])

        if cmds:
            results['changed'] = True
            results['commands'].extend(cmds)

    module.exit_json(**results)


if __name__ == '__main__':
    main()



#
# Copyright (c) 2019 Ericsson AB.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import json

from ansible.module_utils._text import to_text
from ansible.module_utils.basic import env_fallback
from ansible.module_utils.network.common.utils import to_list, ComplexList
from ansible.module_utils.connection import Connection, ConnectionError

_DEVICE_CONFIGS = {}


def get_connection(module):
    if hasattr(module, '_eric_eccli_connection'):
        return module._eric_eccli_connection

    capabilities = get_capabilities(module)
    network_api = capabilities.get('network_api')
    if network_api == 'cliconf':
        module._eric_eccli_connection = Connection(module._socket_path)
    else:
        module.fail_json(msg='Invalid connection type %s' % network_api)

    return module._eric_eccli_connection


def get_capabilities(module):
    if hasattr(module, '_eric_eccli_capabilities'):
        return module._eric_eccli_capabilities
    try:
        capabilities = Connection(module._socket_path).get_capabilities()
    except ConnectionError as exc:
        module.fail_json(msg=to_text(exc, errors='surrogate_then_replace'))
    module._eric_eccli_capabilities = json.loads(capabilities)
    return module._eric_eccli_capabilities


def run_commands(module, commands, check_rc=True):
    connection = get_connection(module)
    try:
        return connection.run_commands(commands=commands, check_rc=check_rc)
    except ConnectionError as exc:
        module.fail_json(msg=to_text(exc))



# coding: utf-8
# Copyright: (c) 2019, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


import asyncio
import datetime
import hashlib

import aiohttp
from jinja2 import Environment, DictLoader


VERSION_FRAGMENT = """
{%- if versions | length > 1 %}
  {% for version in versions %}
    {% if loop.last %}and {{ version }}{% else %}
      {% if versions | length == 2 %}{{ version }} {% else %}{{ version }}, {% endif -%}
    {% endif -%}
  {% endfor -%}
{%- else %}{{ versions[0] }}{% endif -%}
"""

LONG_TEMPLATE = """
{% set plural = False if versions | length == 1 else True %}
{% set latest_ver = (versions | sort(attribute='ver_obj'))[-1] %}

To: ansible-releases@redhat.com, ansible-devel@googlegroups.com, ansible-project@googlegroups.com, ansible-announce@googlegroups.com
Subject: New Ansible release{% if plural %}s{% endif %} {{ version_str }}

{% filter wordwrap %}
Hi all- we're happy to announce that the general release of Ansible {{ version_str }}{% if plural %} are{%- else %} is{%- endif %} now available!
{% endfilter %}



How do you get it?
------------------

{% for version in versions %}
$ pip install ansible=={{ version }} --user
{% if not loop.last %}
or
{% endif %}
{% endfor %}

The tar.gz of the release{% if plural %}s{% endif %} can be found here:

{% for version in versions %}
* {{ version }}
  https://releases.ansible.com/ansible/ansible-{{ version }}.tar.gz
  SHA256: {{ hashes[version] }}
{% endfor %}


What's new in {{ version_str }}
{{ '-' * (14 + version_str | length) }}

{% filter wordwrap %}
{% if plural %}These releases are{% else %}This release is a{% endif %} maintenance release{% if plural %}s{% endif %} containing numerous bugfixes. The full {% if plural %} changelogs are{% else %} changelog is{% endif %} at:
{% endfilter %}


{% for version in versions %}
* {{ version }}
  https://github.com/ansible/ansible/blob/stable-{{ version.split('.')[:2] | join('.') }}/changelogs/CHANGELOG-v{{ version.split('.')[:2] | join('.') }}.rst
{% endfor %}


What's the schedule for future maintenance releases?
----------------------------------------------------

{% filter wordwrap %}
Future maintenance releases will occur approximately every 3 weeks.  So expect the next one around {{ next_release.strftime('%Y-%m-%d') }}.
{% endfilter %}



Porting Help
------------

{% filter wordwrap %}
We've published a porting guide at
https://docs.ansible.com/ansible/devel/porting_guides/porting_guide_{{ latest_ver.split('.')[:2] | join('.') }}.html to help migrate your content to {{ latest_ver.split('.')[:2] | join('.') }}.
{% endfilter %}



{% filter wordwrap %}
If you discover any errors or if any of your working playbooks break when you upgrade to {{ latest_ver }}, please use the following link to report the regression:
{% endfilter %}


  https://github.com/ansible/ansible/issues/new/choose

{% filter wordwrap %}
In your issue, be sure to mention the Ansible version that works and the one that doesn't.
{% endfilter %}


Thanks!

-{{ name }}

"""  # noqa for E501 (line length).
# jinja2 is horrid about getting rid of extra newlines so we have to have a single per paragraph for
# proper wrapping to occur

SHORT_TEMPLATE = """
{% set plural = False if versions | length == 1 else True %}
@ansible
{{ version_str }}
{% if plural %}
  have
{% else %}
  has
{% endif %}
been released! Get
{% if plural %}
them
{% else %}
it
{% endif %}
on PyPI: pip install ansible=={{ (versions|sort(attribute='ver_obj'))[-1] }},
https://releases.ansible.com/ansible/, the Ansible PPA on Launchpad, or GitHub.  Happy automating!
"""  # noqa for E501 (line length).
# jinja2 is horrid about getting rid of extra newlines so we have to have a single per paragraph for
# proper wrapping to occur

JINJA_ENV = Environment(
    loader=DictLoader({'long': LONG_TEMPLATE,
                       'short': SHORT_TEMPLATE,
                       'version_string': VERSION_FRAGMENT,
                       }),
    extensions=['jinja2.ext.i18n'],
    trim_blocks=True,
    lstrip_blocks=True,
)


async def calculate_hash_from_tarball(session, version):
    tar_url = f'https://releases.ansible.com/ansible/ansible-{version}.tar.gz'
    tar_task = asyncio.create_task(session.get(tar_url))
    tar_response = await tar_task

    tar_hash = hashlib.sha256()
    while True:
        chunk = await tar_response.content.read(1024)
        if not chunk:
            break
        tar_hash.update(chunk)

    return tar_hash.hexdigest()


async def parse_hash_from_file(session, version):
    filename = f'ansible-{version}.tar.gz'
    hash_url = f'https://releases.ansible.com/ansible/{filename}.sha'
    hash_task = asyncio.create_task(session.get(hash_url))
    hash_response = await hash_task

    hash_content = await hash_response.read()
    precreated_hash, precreated_filename = hash_content.split(None, 1)
    if filename != precreated_filename.strip().decode('utf-8'):
        raise ValueError(f'Hash file contains hash for a different file: {precreated_filename}')

    return precreated_hash.decode('utf-8')


async def get_hash(session, version):
    calculated_hash = await calculate_hash_from_tarball(session, version)
    precreated_hash = await parse_hash_from_file(session, version)

    if calculated_hash != precreated_hash:
        raise ValueError(f'Hash in file ansible-{version}.tar.gz.sha {precreated_hash} does not'
                         f' match hash of tarball {calculated_hash}')

    return calculated_hash


async def get_hashes(versions):
    hashes = {}
    requestors = {}
    async with aiohttp.ClientSession() as aio_session:
        for version in versions:
            requestors[version] = asyncio.create_task(get_hash(aio_session, version))

        for version, request in requestors.items():
            await request
            hashes[version] = request.result()

    return hashes


def next_release_date(weeks=3):
    days_in_the_future = weeks * 7
    today = datetime.datetime.now()
    numeric_today = today.weekday()

    # We release on Thursdays
    if numeric_today == 3:
        # 3 is Thursday
        pass
    elif numeric_today == 4:
        # If this is Friday, we can adjust back to Thursday for the next release
        today -= datetime.timedelta(days=1)
    elif numeric_today < 3:
        # Otherwise, slide forward to Thursday
        today += datetime.timedelta(days=(3 - numeric_today))
    else:
        # slightly different formula if it's past Thursday this week.  We need to go forward to
        # Thursday of next week
        today += datetime.timedelta(days=(10 - numeric_today))

    next_release = today + datetime.timedelta(days=days_in_the_future)
    return next_release


def create_long_message(versions, name):
    hashes = asyncio.run(get_hashes(versions))

    version_template = JINJA_ENV.get_template('version_string')
    version_str = version_template.render(versions=versions).strip()

    next_release = next_release_date()

    template = JINJA_ENV.get_template('long')
    message = template.render(versions=versions, version_str=version_str,
                              name=name, hashes=hashes, next_release=next_release)
    return message


def create_short_message(versions):
    version_template = JINJA_ENV.get_template('version_string')
    version_str = version_template.render(versions=versions).strip()

    template = JINJA_ENV.get_template('short')
    message = template.render(versions=versions, version_str=version_str)
    message = ' '.join(message.split()) + '\n'
    return message



# (c) 2015, Jonathan Davila <jonathan(at)davila.io>
# (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = """
  lookup: hashi_vault
  author: Jonathan Davila <jdavila(at)ansible.com>
  version_added: "2.0"
  short_description: retrieve secrets from HashiCorp's vault
  requirements:
    - hvac (python library)
  description:
    - retrieve secrets from HashiCorp's vault
  notes:
    - Due to a current limitation in the HVAC library there won't necessarily be an error if a bad endpoint is specified.
  options:
    secret:
      description: query you are making.
      required: True
    token:
      description: vault token.
      env:
        - name: VAULT_TOKEN
    url:
      description: URL to vault service.
      env:
        - name: VAULT_ADDR
      default: 'http://127.0.0.1:8200'
    username:
      description: Authentication user name.
    password:
      description: Authentication password.
    role_id:
      description: Role id for a vault AppRole auth.
      env:
        - name: VAULT_ROLE_ID
    secret_id:
      description: Secret id for a vault AppRole auth.
      env:
        - name: VAULT_SECRET_ID
    auth_method:
      description:
      - Authentication method to be used.
      - C(userpass) is added in version 2.8.
      env:
        - name: VAULT_AUTH_METHOD
      choices:
        - userpass
        - ldap
        - approle
    mount_point:
      description: vault mount point, only required if you have a custom mount point.
      default: ldap
    ca_cert:
      description: path to certificate to use for authentication.
      aliases: [ cacert ]
    validate_certs:
      description: controls verification and validation of SSL certificates, mostly you only want to turn off with self signed ones.
      type: boolean
      default: True
    namespace:
      version_added: "2.8"
      description: namespace where secrets reside. requires HVAC 0.7.0+ and Vault 0.11+.
"""

EXAMPLES = """
- debug:
    msg: "{{ lookup('hashi_vault', 'secret=secret/hello:value token=c975b780-d1be-8016-866b-01d0f9b688a5 url=http://myvault:8200')}}"

- name: Return all secrets from a path
  debug:
    msg: "{{ lookup('hashi_vault', 'secret=secret/hello token=c975b780-d1be-8016-866b-01d0f9b688a5 url=http://myvault:8200')}}"

- name: Vault that requires authentication via LDAP
  debug:
      msg: "{{ lookup('hashi_vault', 'secret=secret/hello:value auth_method=ldap mount_point=ldap username=myuser password=mypas url=http://myvault:8200')}}"

- name: Vault that requires authentication via username and password
  debug:
      msg: "{{ lookup('hashi_vault', 'secret=secret/hello:value auth_method=userpass username=myuser password=mypas url=http://myvault:8200')}}"

- name: Using an ssl vault
  debug:
      msg: "{{ lookup('hashi_vault', 'secret=secret/hola:value token=c975b780-d1be-8016-866b-01d0f9b688a5 url=https://myvault:8200 validate_certs=False')}}"

- name: using certificate auth
  debug:
      msg: "{{ lookup('hashi_vault', 'secret=secret/hi:value token=xxxx-xxx-xxx url=https://myvault:8200 validate_certs=True cacert=/cacert/path/ca.pem')}}"

- name: authenticate with a Vault app role
  debug:
      msg: "{{ lookup('hashi_vault', 'secret=secret/hello:value auth_method=approle role_id=myroleid secret_id=mysecretid url=http://myvault:8200')}}"

- name: Return all secrets from a path in a namespace
  debug:
    msg: "{{ lookup('hashi_vault', 'secret=secret/hello token=c975b780-d1be-8016-866b-01d0f9b688a5 url=http://myvault:8200 namespace=teama/admins')}}"

# to work with kv v2 (vault api - for kv v2 -  GET method requires that PATH should be "secret/data/:path")
- name: Return all kv v2 secrets from a path
  debug:
    msg: "{{ lookup('hashi_vault', 'secret=secret/data/hello token=my_vault_token url=http://myvault_url:8200') }}"


"""

RETURN = """
_raw:
  description:
    - secrets(s) requested
"""

import os

from ansible.errors import AnsibleError
from ansible.module_utils.parsing.convert_bool import boolean
from ansible.plugins.lookup import LookupBase

HAS_HVAC = False
try:
    import hvac
    HAS_HVAC = True
except ImportError:
    HAS_HVAC = False


ANSIBLE_HASHI_VAULT_ADDR = 'http://127.0.0.1:8200'

if os.getenv('VAULT_ADDR') is not None:
    ANSIBLE_HASHI_VAULT_ADDR = os.environ['VAULT_ADDR']


class HashiVault:
    def __init__(self, **kwargs):

        self.url = kwargs.get('url', ANSIBLE_HASHI_VAULT_ADDR)
        self.namespace = kwargs.get('namespace', None)
        self.avail_auth_method = ['approle', 'userpass', 'ldap']

        # split secret arg, which has format 'secret/hello:value' into secret='secret/hello' and secret_field='value'
        s = kwargs.get('secret')
        if s is None:
            raise AnsibleError("No secret specified for hashi_vault lookup")

        s_f = s.rsplit(':', 1)
        self.secret = s_f[0]
        if len(s_f) >= 2:
            self.secret_field = s_f[1]
        else:
            self.secret_field = ''

        self.verify = self.boolean_or_cacert(kwargs.get('validate_certs', True), kwargs.get('cacert', ''))

        # If a particular backend is asked for (and its method exists) we call it, otherwise drop through to using
        # token auth. This means if a particular auth backend is requested and a token is also given, then we
        # ignore the token and attempt authentication against the specified backend.
        #
        # to enable a new auth backend, simply add a new 'def auth_<type>' method below.
        #
        self.auth_method = kwargs.get('auth_method', os.environ.get('VAULT_AUTH_METHOD'))
        self.verify = self.boolean_or_cacert(kwargs.get('validate_certs', True), kwargs.get('cacert', ''))
        if self.auth_method and self.auth_method != 'token':
            try:
                if self.namespace is not None:
                    self.client = hvac.Client(url=self.url, verify=self.verify, namespace=self.namespace)
                else:
                    self.client = hvac.Client(url=self.url, verify=self.verify)
                # prefixing with auth_ to limit which methods can be accessed
                getattr(self, 'auth_' + self.auth_method)(**kwargs)
            except AttributeError:
                raise AnsibleError("Authentication method '%s' not supported."
                                   " Available options are %r" % (self.auth_method, self.avail_auth_method))
        else:
            self.token = kwargs.get('token', os.environ.get('VAULT_TOKEN', None))
            if self.token is None and os.environ.get('HOME'):
                token_filename = os.path.join(
                    os.environ.get('HOME'),
                    '.vault-token'
                )
                if os.path.exists(token_filename):
                    with open(token_filename) as token_file:
                        self.token = token_file.read().strip()

            if self.token is None:
                raise AnsibleError("No Vault Token specified")

            if self.namespace is not None:
                self.client = hvac.Client(url=self.url, token=self.token, verify=self.verify, namespace=self.namespace)
            else:
                self.client = hvac.Client(url=self.url, token=self.token, verify=self.verify)

        if not self.client.is_authenticated():
            raise AnsibleError("Invalid Hashicorp Vault Token Specified for hashi_vault lookup")

    def get(self):
        data = self.client.read(self.secret)

        if data is None:
            raise AnsibleError("The secret %s doesn't seem to exist for hashi_vault lookup" % self.secret)

        if self.secret_field == '':
            return data['data']

        if self.secret_field not in data['data']:
            raise AnsibleError("The secret %s does not contain the field '%s'. for hashi_vault lookup" % (self.secret, self.secret_field))

        return data['data'][self.secret_field]

    def check_params(self, **kwargs):
        username = kwargs.get('username')
        if username is None:
            raise AnsibleError("Authentication method %s requires a username" % self.auth_method)

        password = kwargs.get('password')
        if password is None:
            raise AnsibleError("Authentication method %s requires a password" % self.auth_method)

        mount_point = kwargs.get('mount_point')

        return username, password, mount_point

    def auth_userpass(self, **kwargs):
        username, password, mount_point = self.check_params(**kwargs)
        if mount_point is None:
            mount_point = 'userpass'

        self.client.auth_userpass(username, password, mount_point=mount_point)

    def auth_ldap(self, **kwargs):
        username, password, mount_point = self.check_params(**kwargs)
        if mount_point is None:
            mount_point = 'ldap'

        self.client.auth_ldap(username, password, mount_point=mount_point)

    def boolean_or_cacert(self, validate_certs, cacert):
        validate_certs = boolean(validate_certs, strict=False)
        '''' return a bool or cacert '''
        if validate_certs is True:
            if cacert != '':
                return cacert
            else:
                return True
        else:
            return False

    def auth_approle(self, **kwargs):
        role_id = kwargs.get('role_id', os.environ.get('VAULT_ROLE_ID', None))
        if role_id is None:
            raise AnsibleError("Authentication method app role requires a role_id")

        secret_id = kwargs.get('secret_id', os.environ.get('VAULT_SECRET_ID', None))
        if secret_id is None:
            raise AnsibleError("Authentication method app role requires a secret_id")

        self.client.auth_approle(role_id, secret_id)


class LookupModule(LookupBase):
    def run(self, terms, variables=None, **kwargs):
        if not HAS_HVAC:
            raise AnsibleError("Please pip install hvac to use the hashi_vault lookup module.")

        vault_args = terms[0].split()
        vault_dict = {}
        ret = []

        for param in vault_args:
            try:
                key, value = param.split('=', 1)
            except ValueError:
                raise AnsibleError("hashi_vault lookup plugin needs key=value pairs, but received %s" % terms)
            vault_dict[key] = value

        if 'ca_cert' in vault_dict.keys():
            vault_dict['cacert'] = vault_dict['ca_cert']
            vault_dict.pop('ca_cert', None)

        vault_conn = HashiVault(**vault_dict)

        for term in terms:
            key = term.split()[0]
            value = vault_conn.get()
            ret.append(value)

        return ret



# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The eos l3_interfaces fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from copy import deepcopy
import re

from ansible.module_utils.network.common import utils
from ansible.module_utils.network.eos.argspec.l3_interfaces.l3_interfaces import L3_interfacesArgs


class L3_interfacesFacts(object):
    """ The eos l3_interfaces fact class
    """

    def __init__(self, module, subspec='config', options='options'):
        self._module = module
        self.argument_spec = L3_interfacesArgs.argument_spec
        spec = deepcopy(self.argument_spec)
        if subspec:
            if options:
                facts_argument_spec = spec[subspec][options]
            else:
                facts_argument_spec = spec[subspec]
        else:
            facts_argument_spec = spec

        self.generated_spec = utils.generate_dict(facts_argument_spec)

    def populate_facts(self, connection, ansible_facts, data=None):
        """ Populate the facts for l3_interfaces
        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected configuration
        :rtype: dictionary
        :returns: facts
        """
        if not data:
            data = connection.get('show running-config | section ^interface')

        # split the config into instances of the resource
        resource_delim = 'interface'
        find_pattern = r'(?:^|\n)%s.*?(?=(?:^|\n)%s|$)' % (resource_delim, resource_delim)
        resources = [p.strip() for p in re.findall(find_pattern, data, re.DOTALL)]

        objs = []
        for resource in resources:
            if resource:
                obj = self.render_config(self.generated_spec, resource)
                if obj:
                    objs.append(obj)
        facts = {}
        if objs:
            params = utils.validate_config(self.argument_spec, {'config': objs})
            facts['l3_interfaces'] = [utils.remove_empties(cfg) for cfg in params['config']]
        ansible_facts['ansible_network_resources'].update(facts)
        return ansible_facts

    def render_config(self, spec, conf):
        """
        Render config as dictionary structure and delete keys
          from spec for null values

        :param spec: The facts tree, generated from the argspec
        :param conf: The configuration
        :rtype: dictionary
        :returns: The generated config
        """
        config = deepcopy(spec)

        config['name'] = utils.parse_conf_arg(conf, 'interface')

        matches = re.findall(r'.*ip address (.+)$', conf, re.MULTILINE)
        if matches:
            config["ipv4"] = []
            for match in matches:
                address, dummy, remainder = match.partition(" ")
                ipv4 = {"address": address}
                if remainder == "secondary":
                    ipv4["secondary"] = True
                config['ipv4'].append(ipv4)

        matches = re.findall(r'.*ipv6 address (.+)$', conf, re.MULTILINE)
        if matches:
            config["ipv6"] = []
            for match in matches:
                address, dummy, remainder = match.partition(" ")
                ipv6 = {"address": address}
                config['ipv6'].append(ipv6)

        return utils.remove_empties(config)



# (c) 2015, Ensighten <infra@ensighten.com>
# (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = """
    lookup: credstash
    version_added: "2.0"
    short_description: retrieve secrets from Credstash on AWS
    requirements:
      - credstash (python library)
    description:
      - "Credstash is a small utility for managing secrets using AWS's KMS and DynamoDB: https://github.com/fugue/credstash"
    options:
      _terms:
        description: term or list of terms to lookup in the credit store
        type: list
        required: True
      table:
        description: name of the credstash table to query
        default: 'credential-store'
        required: True
      version:
        description: Credstash version
      region:
        description: AWS region
      profile_name:
        description: AWS profile to use for authentication
        env:
          - name: AWS_PROFILE
      aws_access_key_id:
        description: AWS access key ID
        env:
          - name: AWS_ACCESS_KEY_ID
      aws_secret_access_key:
        description: AWS access key
        env:
          - name: AWS_SECRET_ACCESS_KEY
      aws_session_token:
        description: AWS session token
        env:
          - name: AWS_SESSION_TOKEN
"""

EXAMPLES = """
- name: first use credstash to store your secrets
  shell: credstash put my-github-password secure123

- name: "Test credstash lookup plugin -- get my github password"
  debug: msg="Credstash lookup! {{ lookup('credstash', 'my-github-password') }}"

- name: "Test credstash lookup plugin -- get my other password from us-west-1"
  debug: msg="Credstash lookup! {{ lookup('credstash', 'my-other-password', region='us-west-1') }}"

- name: "Test credstash lookup plugin -- get the company's github password"
  debug: msg="Credstash lookup! {{ lookup('credstash', 'company-github-password', table='company-passwords') }}"

- name: Example play using the 'context' feature
  hosts: localhost
  vars:
    context:
      app: my_app
      environment: production
  tasks:

  - name: "Test credstash lookup plugin -- get the password with a context passed as a variable"
    debug: msg="{{ lookup('credstash', 'some-password', context=context) }}"

  - name: "Test credstash lookup plugin -- get the password with a context defined here"
    debug: msg="{{ lookup('credstash', 'some-password', context=dict(app='my_app', environment='production')) }}"
"""

RETURN = """
  _raw:
    description:
      - value(s) stored in Credstash
"""

import os

from ansible.errors import AnsibleError
from ansible.plugins.lookup import LookupBase

CREDSTASH_INSTALLED = False

try:
    import credstash
    CREDSTASH_INSTALLED = True
except ImportError:
    CREDSTASH_INSTALLED = False


class LookupModule(LookupBase):
    def run(self, terms, variables, **kwargs):

        if not CREDSTASH_INSTALLED:
            raise AnsibleError('The credstash lookup plugin requires credstash to be installed.')

        ret = []
        for term in terms:
            try:
                version = kwargs.pop('version', '')
                region = kwargs.pop('region', None)
                table = kwargs.pop('table', 'credential-store')
                profile_name = kwargs.pop('profile_name', os.getenv('AWS_PROFILE', None))
                aws_access_key_id = kwargs.pop('aws_access_key_id', os.getenv('AWS_ACCESS_KEY_ID', None))
                aws_secret_access_key = kwargs.pop('aws_secret_access_key', os.getenv('AWS_SECRET_ACCESS_KEY', None))
                aws_session_token = kwargs.pop('aws_session_token', os.getenv('AWS_SESSION_TOKEN', None))
                kwargs_pass = {'profile_name': profile_name, 'aws_access_key_id': aws_access_key_id,
                               'aws_secret_access_key': aws_secret_access_key, 'aws_session_token': aws_session_token}
                val = credstash.getSecret(term, version, region, table, context=kwargs, **kwargs_pass)
            except credstash.ItemNotFound:
                raise AnsibleError('Key {0} not found'.format(term))
            except Exception as e:
                raise AnsibleError('Encountered exception while fetching {0}: {1}'.format(term, e))
            ret.append(val)

        return ret



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'core'}


DOCUMENTATION = '''
---
module: gather_facts
version_added: 2.8
short_description: Gathers facts about remote hosts
description:
     - This module takes care of executing the configured facts modules, the default is to use the M(setup) module.
     - This module is automatically called by playbooks to gather useful variables about remote hosts that can be used in playbooks.
     - It can also be executed directly by C(/usr/bin/ansible) to check what variables are available to a host.
     - Ansible provides many I(facts) about the system, automatically.
options:
    parallel:
        description:
            - A toggle that controls if the fact modules are executed in parallel or serially and in order.
              This can guarantee the merge order of module facts at the expense of performance.
            - By default it will be true if more than one fact module is used.
        type: bool
notes:
    - This module is mostly a wrapper around other fact gathering modules.
    - Options passed to this module must be supported by all the underlying fact modules configured.
    - Facts returned by each module will be merged, conflicts will favor 'last merged'.
      Order is not guaranteed, when doing parallel gathering on multiple modules.
author:
    - "Ansible Core Team"
'''

RETURN = """
# depends on the fact module called
"""

EXAMPLES = """
# Display facts from all hosts and store them indexed by I(hostname) at C(/tmp/facts).
# ansible all -m gather_facts --tree /tmp/facts
"""



#!/usr/bin/python
# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
module: ec2_vpc_route_table_info
short_description: Gather information about ec2 VPC route tables in AWS
description:
    - Gather information about ec2 VPC route tables in AWS
    - This module was called C(ec2_vpc_route_table_facts) before Ansible 2.9. The usage did not change.
version_added: "2.0"
author: "Rob White (@wimnat)"
options:
  filters:
    description:
      - A dict of filters to apply. Each dict item consists of a filter key and a filter value.
        See U(https://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_DescribeRouteTables.html) for possible filters.
extends_documentation_fragment:
    - aws
    - ec2
'''

EXAMPLES = '''
# Note: These examples do not set authentication details, see the AWS Guide for details.

# Gather information about all VPC route tables
- ec2_vpc_route_table_info:

# Gather information about a particular VPC route table using route table ID
- ec2_vpc_route_table_info:
    filters:
      route-table-id: rtb-00112233

# Gather information about any VPC route table with a tag key Name and value Example
- ec2_vpc_route_table_info:
    filters:
      "tag:Name": Example

# Gather information about any VPC route table within VPC with ID vpc-abcdef00
- ec2_vpc_route_table_info:
    filters:
      vpc-id: vpc-abcdef00

'''

try:
    import boto.vpc
    from boto.exception import BotoServerError
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.ec2 import AnsibleAWSError, connect_to_aws, ec2_argument_spec, get_aws_connection_info


def get_route_table_info(route_table):

    # Add any routes to array
    routes = []
    associations = []
    for route in route_table.routes:
        routes.append(route.__dict__)
    for association in route_table.associations:
        associations.append(association.__dict__)

    route_table_info = {'id': route_table.id,
                        'routes': routes,
                        'associations': associations,
                        'tags': route_table.tags,
                        'vpc_id': route_table.vpc_id
                        }

    return route_table_info


def list_ec2_vpc_route_tables(connection, module):

    filters = module.params.get("filters")
    route_table_dict_array = []

    try:
        all_route_tables = connection.get_all_route_tables(filters=filters)
    except BotoServerError as e:
        module.fail_json(msg=e.message)

    for route_table in all_route_tables:
        route_table_dict_array.append(get_route_table_info(route_table))

    module.exit_json(route_tables=route_table_dict_array)


def main():
    argument_spec = ec2_argument_spec()
    argument_spec.update(
        dict(
            filters=dict(default=None, type='dict')
        )
    )

    module = AnsibleModule(argument_spec=argument_spec,
                           supports_check_mode=True)
    if module._name == 'ec2_vpc_route_table_facts':
        module.deprecate("The 'ec2_vpc_route_table_facts' module has been renamed to 'ec2_vpc_route_table_info'", version='2.13')

    if not HAS_BOTO:
        module.fail_json(msg='boto required for this module')

    region, ec2_url, aws_connect_params = get_aws_connection_info(module)

    if region:
        try:
            connection = connect_to_aws(boto.vpc, region, **aws_connect_params)
        except (boto.exception.NoAuthHandlerFound, AnsibleAWSError) as e:
            module.fail_json(msg=str(e))
    else:
        module.fail_json(msg="region must be specified")

    list_ec2_vpc_route_tables(connection, module)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 Google
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# ----------------------------------------------------------------------------
#
#     ***     AUTO GENERATED CODE    ***    AUTO GENERATED CODE     ***
#
# ----------------------------------------------------------------------------
#
#     This file is automatically generated by Magic Modules and manual
#     changes will be clobbered when the file is regenerated.
#
#     Please read more about how to change this file at
#     https://www.github.com/GoogleCloudPlatform/magic-modules
#
# ----------------------------------------------------------------------------

from __future__ import absolute_import, division, print_function

__metaclass__ = type

################################################################################
# Documentation
################################################################################

ANSIBLE_METADATA = {'metadata_version': '1.1', 'status': ["preview"], 'supported_by': 'community'}

DOCUMENTATION = '''
---
module: gcp_resourcemanager_project_info
description:
- Gather info for GCP Project
- This module was called C(gcp_resourcemanager_project_facts) before Ansible 2.9.
  The usage has not changed.
short_description: Gather info for GCP Project
version_added: 2.8
author: Google Inc. (@googlecloudplatform)
requirements:
- python >= 2.6
- requests >= 2.18.4
- google-auth >= 1.3.0
options: {}
extends_documentation_fragment: gcp
'''

EXAMPLES = '''
- name: get info on a project
  gcp_resourcemanager_project_info:
    project: test_project
    auth_kind: serviceaccount
    service_account_file: "/tmp/auth.pem"
'''

RETURN = '''
resources:
  description: List of resources
  returned: always
  type: complex
  contains:
    number:
      description:
      - Number uniquely identifying the project.
      returned: success
      type: int
    lifecycleState:
      description:
      - The Project lifecycle state.
      returned: success
      type: str
    name:
      description:
      - 'The user-assigned display name of the Project. It must be 4 to 30 characters.
        Allowed characters are: lowercase and uppercase letters, numbers, hyphen,
        single-quote, double-quote, space, and exclamation point.'
      returned: success
      type: str
    createTime:
      description:
      - Time of creation.
      returned: success
      type: str
    labels:
      description:
      - The labels associated with this Project.
      - 'Label keys must be between 1 and 63 characters long and must conform to the
        following regular expression: `[a-z]([-a-z0-9]*[a-z0-9])?`.'
      - Label values must be between 0 and 63 characters long and must conform to
        the regular expression `([a-z]([-a-z0-9]*[a-z0-9])?)?`.
      - No more than 256 labels can be associated with a given resource.
      - Clients should store labels in a representation such as JSON that does not
        depend on specific characters being disallowed .
      returned: success
      type: dict
    parent:
      description:
      - A parent organization.
      returned: success
      type: complex
      contains:
        type:
          description:
          - Must be organization.
          returned: success
          type: str
        id:
          description:
          - Id of the organization.
          returned: success
          type: str
    id:
      description:
      - The unique, user-assigned ID of the Project. It must be 6 to 30 lowercase
        letters, digits, or hyphens. It must start with a letter.
      - Trailing hyphens are prohibited.
      returned: success
      type: str
'''

################################################################################
# Imports
################################################################################
from ansible.module_utils.gcp_utils import navigate_hash, GcpSession, GcpModule, GcpRequest
import json

################################################################################
# Main
################################################################################


def main():
    module = GcpModule(argument_spec=dict())

    if module._name == 'gcp_resourcemanager_project_facts':
        module.deprecate("The 'gcp_resourcemanager_project_facts' module has been renamed to 'gcp_resourcemanager_project_info'", version='2.13')

    if not module.params['scopes']:
        module.params['scopes'] = ['https://www.googleapis.com/auth/cloud-platform']

    return_value = {'resources': fetch_list(module, collection(module))}
    module.exit_json(**return_value)


def collection(module):
    return "https://cloudresourcemanager.googleapis.com/v1/projects".format(**module.params)


def fetch_list(module, link):
    auth = GcpSession(module, 'resourcemanager')
    return auth.list(link, return_if_object, array_name='projects')


def return_if_object(module, response):
    # If not found, return nothing.
    if response.status_code == 404:
        return None

    # If no content, return nothing.
    if response.status_code == 204:
        return None

    try:
        module.raise_for_status(response)
        result = response.json()
    except getattr(json.decoder, 'JSONDecodeError', ValueError) as inst:
        module.fail_json(msg="Invalid JSON response with error: %s" % inst)

    if navigate_hash(result, ['error', 'errors']):
        module.fail_json(msg=navigate_hash(result, ['error', 'errors']))

    return result


if __name__ == "__main__":
    main()
